
from logging import getLogger

from ..lib.date import to_time
from ..sql import ActivityBookmark, ActivityGroup, StatisticName
from ..sql.database import connect
from ..stats.read.segment import SegmentReader


log = getLogger(__name__)


class CoastingBookmark:

    def __init__(self, db):
        self._db = db

    def run(self, kit, min_time, max_cadence, min_speed, group, constraint=None):
        with self._db.session_context() as s:
            self.__delete_previous(s, constraint)
            for aj_id, start, finish in self.__find(s, kit, min_time, max_cadence, min_speed, group):
                self.__add(s, aj_id, start, finish, constraint)

    def __delete_previous(self, s, constraint):
        s.query(ActivityBookmark). \
            filter(ActivityBookmark.owner == self,
                   ActivityBookmark.constraint == constraint). \
            delete()

    def __add(self, s, aj_id, start, finish, constraint):
        log.info('%s - %s (%d)' % (to_time(start), to_time(finish), aj_id))
        s.add(ActivityBookmark(activity_journal_id=aj_id, start=start, finish=finish,
                               owner=self, constraint=constraint))

    def __find(self, s, kit, min_time, max_cadence, min_speed, group):
        group = ActivityGroup.from_name(s, group)
        cadence_statistic = StatisticName.from_name(s, 'Cadence', SegmentReader, group)
        distance_statistic = StatisticName.from_name(s, 'Distance', SegmentReader, group)
        speed_statistic = StatisticName.from_name(s, 'Speed', SegmentReader, group)
        kit_statistic = StatisticName.from_name(s, 'kit', SegmentReader, group)
        s.execute('PRAGMA automatic_index=OFF;')
        for row in s.execute('''
select t.activity_journal_id, s.time, f.time
  from statistic_journal as ks,
       statistic_journal_text as kt,
       activity_timespan as t cross join  -- this tweaked the plan
       statistic_journal as s,
       statistic_journal_integer as si,
       statistic_journal as ss,
       statistic_journal_integer as ssi,
       statistic_journal as f,
       statistic_journal_integer as fi,
       statistic_journal as ff,
       statistic_journal_integer as ffi
 where ks.source_id = t.activity_journal_id
   and ks.statistic_name_id = :kit_statistic
   and ks.id = kt.id
   and kt.value = :kit
   -- s and ss bracket the change in cadence at the start
   -- f and ff bracket the change in cadence at the finish
   and s.statistic_name_id = :cadence_statistic
   and f.statistic_name_id = :cadence_statistic
   and ss.statistic_name_id = :cadence_statistic
   and ff.statistic_name_id = :cadence_statistic
   and s.id = si.id
   and f.id = fi.id
   and ss.id = ssi.id
   and ff.id = ffi.id
   and si.value < :max_cadence
   and fi.value < :max_cadence
   and ss.serial = s.serial-1
   and ff.serial = f.serial+1
   and s.source_id = t.activity_journal_id
   and ss.source_id = t.activity_journal_id
   and f.source_id = t.activity_journal_id
   and ff.source_id = t.activity_journal_id
   and ssi.value >= :max_cadence
   and ffi.value >= :max_cadence
   and f.time - s.time > :min_time
   and t.start < s.time
   and t.finish > f.time
   -- limit cadence between s and f
   and not exists (select 1
                     from statistic_journal_integer as ji,
                          statistic_journal as j
                    where ji.value >= :max_cadence
                      and ji.id = j.id
                      and j.statistic_name_id = :cadence_statistic
                      and j.source_id = s.source_id
                      and j.serial > s.serial
                      and j.serial < f.serial)
   -- limit average speed
   and exists (select 1
                 from statistic_journal_float as f1,
                      statistic_journal_float as f2,
                      statistic_journal as j1,
                      statistic_journal as j2
                where f1.id = j1.id
                  and f2.id = j2.id
                  and j1.statistic_name_id = :distance_statistic
                  and j2.statistic_name_id = :distance_statistic
                  and j1.serial = s.serial
                  and j2.serial = f.serial
                  and j1.source_id = f.source_id
                  and j2.source_id = f.source_id
                  and f2.value - f1.value > :min_speed * (f.time - s.time))
   -- exclude stationary points 
   and not exists (select 1
                     from statistic_journal_float as jf,
                          statistic_journal as j
                    where jf.value = 0
                      and jf.id = j.id
                      and j.statistic_name_id = :speed_statistic
                      and j.source_id = s.source_id
                      and j.serial >= s.serial
                      and j.serial <= f.serial)
 order by t.activity_journal_id;  -- doesn't affect speed and makes duplicates easier to see
        ''', {'min_time': min_time, 'max_cadence': max_cadence, 'min_speed': min_speed, 'kit': kit,
              'group': str(group), 'cadence_statistic': cadence_statistic.id, 'speed_statistic': speed_statistic.id,
              'distance_statistic': distance_statistic.id, 'kit_statistic': kit_statistic.id}):
            yield row[0], row[1], row[2]
        s.execute('PRAGMA automatic_index=ON;')


if __name__ == '__main__':
    '''
    Bookmark sections of activities that exceed 60s of little or no pedalling (maximum cadence 20)
    and with any speed (max 0).
    
    These are then used in the fit_power_parameters template to estimate CdA and Crr. 
    '''
    ns, db = connect(['-v 4'])
#    CoastingBookmark(db).run('cotic', 60, 20, 0, 'bike', constraint='cotic 60/20/0')
    CoastingBookmark(db).run('cotic', 15, 5, 10, 'bike', constraint='cotic 15/5/10')




from bokeh.io import output_notebook, show
from math import pi, sin, cos

from ch2.data import *
from ch2.data.plot import *
from ch2.jupyter.decorator import template
from ch2.sql import *


@template
def nearby_activities(constraint):

    f'''
    # Nearby Activities: {constraint}
    '''

    '''
    $contents
    '''

    '''
    ## Load Group
    '''
    s = session('-v2')
    groups = df(s.query(ActivityNearby).filter(ActivityNearby.constraint == constraint))
    n_groups = len(groups.constraint.unique())

    '''
    ## Load and Plot Route Data
    '''

    offset_a, offset_b = 400, 150
    x_max, y_max = None, None
    # you can provide manual labels here
    labels = None
    palette = list(evenly_spaced_hues(n_groups, saturation=0.8, value=0.8))

    output_notebook()
    f = figure(plot_width=900, plot_height=900, x_axis_type='mercator', y_axis_type='mercator')
    f.add_tile(STAMEN_TERRAIN, alpha=0.5)

    for i in range(n_groups):
        source_ids = groups.loc[groups.constraint == i].activity_journal_id
        n_source_ids = len(source_ids)
        a = 2 * pi * i / n_groups
        dxa, dya = offset_a * sin(a), offset_a * cos(a)
        for j, source_id in enumerate(source_ids):
            b = 2 * pi * j / n_source_ids
            dxb, dyb = offset_b * sin(b), offset_b * cos(b)
            stats = activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                     activity_journal=ActivityJournal.from_id(s, source_id))
            f.line(x=stats[SPHERICAL_MERCATOR_X] + dxa + dxb, y=stats[SPHERICAL_MERCATOR_Y] + dya + dyb,
                   color=palette[i], line_width=1.5, line_dash='dotted')
            if x_max is None:
                x_max, y_max = max(stats[SPHERICAL_MERCATOR_X]), max(stats[SPHERICAL_MERCATOR_Y])
            else:
                x_max, y_max = max(x_max, max(stats[SPHERICAL_MERCATOR_X])), max(y_max, max(stats[SPHERICAL_MERCATOR_Y]))

    if not labels:
        labels = [f'{constraint} {g}' for g in range(n_groups)]
    for i in range(n_groups):
        source_ids = groups.loc[groups.constraint == i].activity_journal_id
        n_source_ids = len(source_ids)
        f.line(x=[x_max, x_max * 1.001], y=[y_max, y_max], color=palette[i], line_width=5, line_dash='solid',
               legend_label='%s (%d)' % (labels[i], n_source_ids))

    f.xaxis.axis_label = 'Longitude'
    f.yaxis.axis_label = 'Latitude'
    show(f)




from io import BytesIO
from logging import getLogger
from os.path import join, exists
from re import search, sub
from zipfile import ZipFile

from requests import session

log = getLogger(__name__)


'''
if you want to download your data from garmin you do not need to understand this code.  install choochoo and 
use the 'ch2 garmin' command.

BUT there is an easier way.  simply go to garmin and ask for all your data.  follow the instructions at

  https://support.garmin.com/en-US/?faq=W1TvTPW8JZ6LfJSfK512Q8

(the second set of instructions is for everything).

you can also go to

  https://www.garmin.com/en-US/account/datamanagement/

and (once logged in) click on 'export your data'.  this gives you data in the raw format (fit files).

i would strongly encourage you to use those approaches rather than my code.  it will be simpler and more reliable.
'''


class GarminConnect:

    # logic and data from https://github.com/tcgoetz/GarminDB/blob/master/download_garmin.py
    # and also a bunch more similar scripts

    base_url = 'https://connect.garmin.com'
    sso_url = 'https://sso.garmin.com/sso'
    modern = base_url + '/modern'
    signin = sso_url + '/signin'
    daily = modern + '/proxy/download-service/files/wellness'

    headers = {
        # 'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/62.0'
        'User-Agent': 'Mozilla/5.0 (compatible; https://github.com/andrewcooke/choochoo)',
        'origin': 'https://sso.garmin.com'
    }

    def __init__(self, log_response=False):
        self._r = session()
        self._log_response = log_response

    def login(self, username, password):

        log.info('Connecting to Garmin Connect as %s' % username)

        params = {
            # todo - are these all actually needed?  by an sso service?
            'webhost': self.base_url,
            'service': self.modern,
            'source': self.signin,
            'redirectAfterAccountLoginUrl': self.modern,
            'redirectAfterAccountCreationUrl': self.modern,
            'gauthHost': self.sso_url,
            'locale': 'en_US',
            'id': 'gauth-widget',
            'cssUrl': 'https://static.garmincdn.com/com.garmin.connect/ui/css/gauth-custom-v1.2-min.css',
            'clientId': 'GarminConnect',
            'rememberMeShown': 'true',
            'rememberMeChecked': 'false',
            'createAccountShown': 'true',
            'openCreateAccount': 'false',
            'usernameShown': 'false',
            'displayNameShown': 'false',
            'consumeServiceTicket': 'false',
            'initialFocus': 'true',
            'embedWidget': 'false',
            'generateExtraServiceTicket': 'false'
        }

        response = self._log_r(self._r.get(self.signin, headers=self.headers, params=params))
        response.raise_for_status()

        data = {
            'username': username,
            'password': password,
            'embed': 'true',
            'lt': 'e1s1',
            '_eventId': 'submit',
            'displayNameRequired': 'false'
        }

        response = self._log_r(self._r.post(self.signin, headers=self.headers, params=params, data=data))
        response.raise_for_status()

        response_url = search(r'"(https:[^"]+?ticket=[^"]+)"', response.text)
        if not response_url:
            log.debug(response.text)
            raise Exception('Could not find response URL')
        response_url = sub(r'\\', '', response_url.group(1))
        log.debug('Response URL: %s' % response_url)

        response = self._log_r(self._r.get(response_url))
        response.raise_for_status()

    def get_monitoring(self, date):
        response = self._log_r(self._r.get(self.daily + date.strftime('/%Y-%m-%d'), headers=self.headers))
        response.raise_for_status()
        return response

    def _log_r(self, response):
        if self._log_response:
            log.debug('headers: %s' % response.headers)
            log.debug('reason: %s' % response.reason)
            log.debug('cookies: %s' % response.cookies)
            log.debug('history: %s' % response.history)
            # log.debug('text: %s' % response.text)
        return response

    def get_monitoring_to_zip_file(self, date, dir):
        path = join(dir, date.strftime('%Y-%m-%d.zip'))
        if exists(path):
            raise Exception('"%s" already exists' % path)
        response = self.get_monitoring(date)
        with open(path, 'wb') as f:
            f.write(response.content)
        log.info('Downloaded data for %s to %s' % (date, path))

    def get_monitoring_to_fit_file(self, date, data_dir, old_format=False):
        from ...commands.upload import write_file, DATA, hash_file, NAME, parse_fit_data, build_path
        response = self.get_monitoring(date)
        zipfile = ZipFile(BytesIO(response.content))
        for name in zipfile.namelist():
            if old_format:
                path = zipfile.extract(name, path=data_dir)
                log.info('Downloaded data for %s to %s' % (date, path))
            else:
                file = {DATA: zipfile.read(name), NAME: name}
                hash_file(file)
                parse_fit_data(file)
                build_path(data_dir, file)
                write_file(file)




from collections import defaultdict
from logging import getLogger

from sqlalchemy import Column, Integer, ForeignKey, Text, desc, or_
from sqlalchemy.orm import relationship, aliased, backref
from sqlalchemy.orm.exc import NoResultFound

from .source import Source, SourceType, Composite, CompositeComponent
from .statistic import StatisticJournal, StatisticName, StatisticJournalTimestamp
from ..support import Base
from ..utils import add
from ...commands.args import FORCE, mm
from ...diary.model import TYPE, DB, UNITS, VALUE
from ...lib import now, time_to_local_time
from ...lib.date import YMD
from ...lib.utils import inside_interval
from ...stats.names import KIT_ADDED, KIT_RETIRED, KIT_USED, ACTIVE_TIME, ACTIVE_DISTANCE, KM, S, _s, AGE, D

log = getLogger(__name__)

NAME = 'name'
GROUP = 'group'
ITEM = 'item'
ITEMS = _s(ITEM)
COMPONENT = 'component'
COMPONENTS = _s(COMPONENT)
MODEL = 'model'
MODELS = _s(MODEL)
STATISTICS = 'statistics'
N = 'n'
SUM = 'sum'
MEAN = 'mean'
MEDIAN = 'median'
ADDED = 'added'
EXPIRED = 'expired'

# statistics calculations
INDIVIDUAL = 'individual'
POPULATION = 'population'


'''
difficult design decisions here. 
  * too complex for integration with constants.
  * trade-off between simplicity and structure.  no type for top-level items, for example.
  * all parts are automatically given statistics for start and finish times
  * tie-in with activities to get active time / distance
in the end, what drove the design was the commands (see commands/kit.py) - trying to keep them as simple as possible.
unfortunately that pushed some extra complexity into the data model (eg to guarantee all names unique).
'''


def get_name(s, name, classes=None, require=False):
    # these cannot be parameter defaults because they're undefined at the top level
    classes = classes or (KitGroup, KitItem, KitComponent, KitModel)
    for cls in classes:
        # can be multiple models, in which case we return one 'at random'
        instance = s.query(cls).filter(cls.name == name).first()
        if instance:
            return instance
    if require:
        raise Exception(f'Cannot find "{name}"')


def assert_name_does_not_exist(s, name, use):
    instance = get_name(s, name)
    if instance and not isinstance(instance, use):
        raise Exception(f'The name "{name}" is already used for a {type(instance).SIMPLE_NAME}')


def expand_item(s, name, time):
    try:
        item = s.query(KitItem).filter(KitItem.name == name).one()
        start, finish = item.time_added(s), item.time_expired(s)
        if start <= time and (not finish or finish >= time):
            log.debug(f'Found {item}')
            yield item
            for model in s.query(KitModel).filter(KitModel.item == item).all():
                start, finish = model.time_added(s), model.time_expired(s)
                if start <= time and (not finish or finish >= time):
                    log.debug(f'Found {model}')
                    yield model
                else:
                    log.debug(f'Outside time {start} <= {time} <= {finish}')
        else:
            log.debug(f'Outside time {start} <= {time} <= {finish}')
    except NoResultFound as e:
        log.error(e)
        raise Exception(f'Kit item {name} not defined')


class StatisticsMixin:
    '''
    support for reading and writing statistics related to a kit class.
    '''

    def _base_statistic_query(self, s, statistic, *sources, owner=None):
        sources = (self,) + sources
        subq = s.query(Composite.id.label('composite_id'))
        for source in sources:
            cc = aliased(CompositeComponent)
            subq = subq.join(cc, Composite.id == cc.output_source_id).filter(cc.input_source == source)
        subq = subq.subquery()
        q = s.query(StatisticJournal). \
            join(StatisticName). \
            outerjoin(subq, subq.c.composite_id == StatisticJournal.source_id). \
            filter(StatisticName.name == statistic,
                   StatisticName.constraint == None)
        if len(sources) == 1:
            q = q.filter(or_(StatisticJournal.source == self, subq.c.composite_id != None))
        else:
            q = q.filter(subq.c.composite_id != None)
        if owner:
            q = q.filter(StatisticName.owner == owner)
        return q

    def _base_use_query(self, s, statistic):
        cc1, cc2 = aliased(CompositeComponent), aliased(CompositeComponent)
        sourceq = s.query(cc1.input_source_id). \
            join(Composite, Composite.id == cc1.output_source_id). \
            join(cc2, cc2.output_source_id == Composite.id). \
            filter(cc2.input_source == self,
                   cc1.input_source != self).subquery()
        return s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == statistic,
                   StatisticJournal.source_id.in_(sourceq))

    def _get_statistic(self, s, statistic, *sources, owner=None):
        return self._base_statistic_query(s, statistic, *sources, owner=owner).one_or_none()

    def _get_statistics(self, s, statistic, *sources, owner=None):
        return self._base_statistic_query(s, statistic, *sources, owner=owner).all()

    def _remove_statistic(self, s, statistic, *sources, owner=None):
        # cannot delete directly with join
        for instance in self._get_statistics(s, statistic, *sources, owner=owner):
            s.delete(instance)

    def _add_timestamp(self, s, statistic, time, source=None, owner=None):
        log.debug(f'Add timestamp for {statistic} at {time} with source {source}')
        # if source is given it is in addition to self
        if source:
            source = Composite.create(s, source, self)
            log.debug(f'Composite {source}')
        else:
            source = self
        owner = owner or self
        return StatisticJournalTimestamp.add(s, statistic, None, None, owner, None, source, time)

    def time_added(self, s):
        return self._get_statistic(s, KIT_ADDED).time

    def time_expired(self, s):
        try:
            return self._get_statistic(s, KIT_RETIRED).time
        except AttributeError:
            return None

    def add_use(self, s, time, source=None, owner=None):
        self._add_timestamp(s, KIT_USED, time, source=source, owner=owner)

    def active_times(self, s):
        return self._base_use_query(s, ACTIVE_TIME).all()

    def active_distances(self, s):
        return self._base_use_query(s, ACTIVE_DISTANCE).all()

    def lifetime(self, s):
        added, expired = self.time_added(s), self.time_expired(s)
        expired = expired or now()
        return expired - added

    def _add_individual_statistics(self, s, model):
        model_statistics = []
        self._calculate_individual_statistics(model_statistics, ACTIVE_DISTANCE, self.active_distances(s), KM)
        self._calculate_individual_statistics(model_statistics, ACTIVE_TIME, self.active_times(s), S)
        expire = self.time_expired(s) or now()
        model_statistics.append({NAME: AGE, N: 1, SUM: (expire - self.time_added(s)).days, UNITS: D})
        model[STATISTICS] = model_statistics

    def _calculate_individual_statistics(self, model_statistics, name, values, units):
        n = len(values)
        if n:
            values = [value.value for value in values]
            total = sum(values)
            # had mean and median, but they were pointless
            model_statistics.append({NAME: name, N: n, SUM: total, UNITS: units})


class ModelMixin:

    @staticmethod
    def fmt_time(time):
        if time:
            return time_to_local_time(time, fmt=YMD)
        else:
            return None

    def to_model(self, s, depth=0, statistics=None, time=None, own_models=True):
        model = {TYPE: self.SIMPLE_NAME, DB: self.id, NAME: self.name}
        try:
            model.update({ADDED: self.fmt_time(self.time_added(s)),
                          EXPIRED: self.fmt_time(self.time_expired(s))})
        except AttributeError:
            pass  # not a subclass of statistics mixin
        if depth > 0:
            self._add_children(s, model, depth=depth-1, statistics=statistics, time=time, own_models=own_models)
        try:
            if statistics == INDIVIDUAL:
                self._add_individual_statistics(s, model)
        except AttributeError:
            log.debug(f'No {statistics} statistics for {self.SIMPLE_NAME} {self.name}')
        return model


class KitGroup(ModelMixin, Base):
    '''
    top level group for kit (bike, shoe, etc)
    '''

    __tablename__ = 'kit_group'
    SIMPLE_NAME = 'group'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, index=True, unique=True)

    @classmethod
    def get_or_add(cls, s, name, force=False):
        try:
            return s.query(KitGroup).filter(KitGroup.name == name).one()
        except NoResultFound:
            assert_name_does_not_exist(s, name, KitGroup)
            if force:
                log.warning(f'Forcing creation of new group ({name})')
                return add(s, KitGroup(name=name))
            else:
                groups = \
                    s.query(KitGroup).order_by(KitGroup.name).all()
                if groups:
                    log.info('Existing groups:')
                    for existing in groups:
                        log.info(f'  {existing.name}')
                    raise Exception(f'Give an existing group, or specify {mm(FORCE)} to create a new group ({name})')
                else:
                    raise Exception(f'Specify {mm(FORCE)} to create a new group ({name})')

    def _add_children(self, s, model, depth=0, statistics=None, time=None, own_models=True):
        model[ITEMS] = [item.to_model(s, depth=depth, statistics=statistics, 
                                      time=time, own_models=own_models)
                        for item in self.items
                        if time is None or inside_interval(item.time_added(s), time, item.time_expired(s))]

    def __str__(self):
        return f'KitGroup "{self.name}"'


class KitItem(ModelMixin, StatisticsMixin, Source):
    '''
    an individual kit item (a particular bike, a particular shoe, etc)
    '''

    __tablename__ = 'kit_item'
    SIMPLE_NAME = 'item'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    group_id = Column(Integer, ForeignKey('kit_group.id', ondelete='cascade'), nullable=False, index=True)
    group = relationship('KitGroup', backref=backref('items', passive_deletes=True))
    name = Column(Text, nullable=False, index=True, unique=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ITEM
    }

    @classmethod
    def add(cls, s, group, name, date):
        # don't rely on unique index to catch duplicates because that's not triggered until commit
        if s.query(KitItem).filter(KitItem.name == name).count():
            raise Exception(f'Item {name} of group {group.name} already exists')
        else:
            assert_name_does_not_exist(s, name, KitItem)
            item = add(s, KitItem(group=group, name=name))
            item._add_statistics(s, date)
            return item

    def _add_statistics(self, s, time):
        self._add_timestamp(s, KIT_ADDED, time)

    @classmethod
    def get(cls, s, name):
        try:
            return s.query(KitItem).filter(KitItem.name == name).one()
        except NoResultFound:
            raise Exception(f'Item {name} does not exist')

    @property
    def components(self):
        components = defaultdict(list)
        for model in self.models:
            components[model.component].append(model)
        return components

    def finish(self, s, date, force):
        if self.time_expired(s):
            if force:
                self._remove_statistic(s, KIT_RETIRED, owner=self)
            else:
                raise Exception(f'Item {self.name} is already retired')
        self._add_timestamp(s, KIT_RETIRED, date)

    def delete(self, s):
        s.delete(self)
        Composite.clean(s)

    def _add_children(self, s, model, depth=0, statistics=None, time=None, own_models=True):
        model[COMPONENTS] = [component.to_model(s, depth=depth, statistics=statistics, 
                                                time=time, own_models=own_models)
                             for component in self.components]
        model[MODELS] = [model.to_model(s, depth=depth, statistics=statistics, 
                                        time=time, own_models=own_models)
                         for model in self.models
                         if time is None or inside_interval(model.time_added(s), time, model.time_expired(s))]

    def to_model(self, s, depth=0, statistics=None, time=None, own_models=True):
        model = super().to_model(s, depth=depth, statistics=statistics, time=time)
        model_ids = set(model.id for model in self.models)
        if own_models and COMPONENTS in model:
            for component in model[COMPONENTS]:
                # restrict component's models to subset of own models
                component[MODELS] = [model for model in component[MODELS] if model[DB] in model_ids]
        return model

    def __str__(self):
        return f'KitItem "{self.name}"'


class KitComponent(ModelMixin, Base):
    '''
    the kind of thing that a kit item is made of (bike wheel, shoe laces, etc)
    '''

    __tablename__ = 'kit_component'
    SIMPLE_NAME = 'component'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, index=True)

    @classmethod
    def get(cls, s, name, require=True):
        instance = s.query(KitComponent).filter(KitComponent.name == name).one_or_none()
        if require and not instance:
            raise Exception(f'Component {name} does not exist')
        return instance

    @classmethod
    def get_or_add(cls, s, name, force):
        try:
            return s.query(KitComponent).filter(KitComponent.name == name).one()
        except NoResultFound:
            assert_name_does_not_exist(s, name, KitComponent)
            if force:
                log.warning(f'Forcing creation of new component ({name})')
                return add(s, KitComponent(name=name))
            else:
                components = s.query(KitComponent).order_by(KitComponent.name).all()
                if components:
                    log.info('Existing components:')
                    for existing in components:
                        log.info(f'  {existing.name}')
                    raise Exception(f'Give an existing component, or specify {mm(FORCE)} to create a new one ({name})')
                else:
                    raise Exception(f'Specify {mm(FORCE)} to create a new component ({name})')

    def delete_if_unused(self, s):
        s.refresh(self)  # make sure deleted models are no longer present
        if not self.models:
            s.delete(self)

    def _add_children(self, s, model, depth=0, statistics=None, time=None, own_models=True):
        model_statistics = INDIVIDUAL if statistics == POPULATION else statistics
        model[MODELS] = [model.to_model(s, depth=depth, statistics=model_statistics,
                                        time=time, own_models=own_models)
                         for model in self.models
                         if time is None or inside_interval(model.time_added(s), time, model.time_expired(s))]
        if statistics == POPULATION:
            self._add_population_statistics(model)

    def to_model(self, s, depth=0, statistics=None, time=None, own_models=True):
        # force all time, since this is constrained via the item if needed and, if not, helps prompts
        return super().to_model(s, depth=depth, statistics=statistics, time=time if statistics else None)

    def _add_population_statistics(self, model):
        # replace KitModel instances with populations.  so this doesn't just rewrite the statistics,
        # it can actually rewrite the model
        population_models = {}
        for instance in model[MODELS]:
            name = instance[NAME]
            if name not in population_models:
                population_models[name] = instance
                for statistic in instance[STATISTICS]:
                    statistic[N] = 1
                    statistic[MEAN] = statistic[SUM]
                    del statistic[SUM]
                del instance[DB]
                del instance[ADDED]
                del instance[EXPIRED]
            else:
                population = population_models[NAME]
                for instance_statistic in instance[STATISTICS]:
                    population_statistic = population[STATISTICS]. \
                        filter(lambda statistic: statistic[NAME] == instance_statistic[NAME])[0]
                    population_statistic[N] += 1
                    population_statistic[MEAN] += instance_statistic[SUM]
        model[MODELS] = []
        for name in population_models:
            population = population_models[name]
            for statistic in population[STATISTICS]:
                statistic[MEAN] /= statistic[N]
            model[MODELS].append(population)

    def __str__(self):
        return f'KitComponent "{self.name}"'


class KitModel(ModelMixin, StatisticsMixin, Source):
    '''
    a particular piece of a kit item (a particular bike wheel, a particular set of laces, etc).
    '''

    __tablename__ = 'kit_model'
    SIMPLE_NAME = 'model'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    item_id = Column(Integer, ForeignKey('kit_item.id', ondelete='cascade'), nullable=False, index=True)
    item = relationship('KitItem', foreign_keys=[item_id], backref=backref('models', passive_deletes=True))
    component_id = Column(Integer, ForeignKey('kit_component.id', ondelete='cascade'), nullable=False, index=True)
    component = relationship('KitComponent', backref=backref('models', passive_deletes=True))
    name = Column(Text, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.MODEL
    }

    @classmethod
    def add(cls, s, item, component, name, time):
        cls._reject_duplicate(s, item, component, name, time)
        model = cls._add_instance(s, item, component, name)
        model._add_statistics(s, time)
        return model

    @classmethod
    def _reject_duplicate(cls, s, item, component, name, time):
        if s.query(StatisticJournal). \
                join(StatisticName). \
                join(KitModel, KitModel.id == StatisticJournal.source_id). \
                filter(StatisticName.name == KIT_ADDED,
                       StatisticJournal.time == time,
                       KitModel.name == name,
                       KitModel.component == component,
                       KitModel.item == item).count():
            raise Exception(f'This part already exists at this date')

    @classmethod
    def _add_instance(cls, s, item, component, name):
        # TODO - restrict name to a particular component
        if not s.query(KitModel).filter(KitModel.name == name).count():
            assert_name_does_not_exist(s, name, KitModel)
            log.warning(f'Model {name} does not match any previous entries')
        return add(s, KitModel(item=item, component=component, name=name))

    def _add_statistics(self, s, time):
        self._add_timestamp(s, KIT_ADDED, time)
        before = self.before(s, time)
        after = self.after(s, time)
        if before:
            before_expiry = before.time_expired(s)
            if before_expiry and before_expiry > time:
                before._remove_statistic(s, KIT_RETIRED)
                before_expiry = None
            if not before_expiry:
                before._add_timestamp(s, KIT_RETIRED, time)
                log.info(f'Retired previous {self.component.name} ({before.name})')
        if after:
            after_added = after.time_added(s)
            self._add_timestamp(s, KIT_RETIRED, after_added)
            log.info(f'Retired new {self.component.name} ({self.name})')

    @classmethod
    def get_all_at(cls, s, item, time):
        beforeq = s.query(StatisticJournalTimestamp.source_id, StatisticJournalTimestamp.time). \
            join(StatisticName). \
            filter(StatisticName.name == KIT_ADDED).subquery()
        afterq = s.query(StatisticJournalTimestamp.source_id, StatisticJournalTimestamp.time). \
            join(StatisticName). \
            filter(StatisticName.name == KIT_RETIRED).subquery()
        return s.query(KitModel). \
            join(beforeq, beforeq.c.source_id == KitModel.id). \
            outerjoin(afterq, afterq.c.source_id == KitModel.id). \
            filter(KitModel.item == item,
                   beforeq.c.time <= time,
                   or_(afterq.c.time >= time, afterq.c.time == None)).all()

    @classmethod
    def get_all(cls, s, item, component):
        return s.query(KitModel). \
            filter(KitModel.item == item,
                   KitModel.component == component).all()

    @classmethod
    def get(cls, s, item, component, name, time, require=True):
        # if time is None, any appropriate model is returned
        q = s.query(KitModel)
        if time:
            beforeq = s.query(StatisticJournalTimestamp.source_id, StatisticJournalTimestamp.time). \
                join(StatisticName). \
                filter(StatisticName.name == KIT_ADDED).subquery()
            afterq = s.query(StatisticJournalTimestamp.source_id, StatisticJournalTimestamp.time). \
                join(StatisticName). \
                filter(StatisticName.name == KIT_RETIRED).subquery()
            q = q.join(beforeq, beforeq.c.source_id == KitModel.id). \
                outerjoin(afterq, afterq.c.source_id == KitModel.id). \
                filter(beforeq.c.time <= time,
                       or_(afterq.c.time >= time, afterq.c.time == None))
        instance = q.filter(KitModel.item == item,
                            KitModel.component == component,
                            KitModel.name == name).first()
        if not instance and require:
            raise Exception(f'Model {name} does not exist')
        return instance

    def _base_sibling_query(self, s, statistic):
        return s.query(KitModel). \
            join(StatisticJournal, StatisticJournal.source_id == KitModel.id). \
            join(StatisticName). \
            join(KitComponent, KitComponent.id == KitModel.component_id). \
            join(KitItem, KitItem.id == KitModel.item_id). \
            filter(StatisticName.name == statistic,
                   KitComponent.name == self.component.name,
                   KitItem.name == self.item.name)

    def before(self, s, time=None):
        if not time:
            time = self.time_added(s)
        return self._base_sibling_query(s, KIT_ADDED).filter(StatisticJournal.time < time). \
            order_by(desc(StatisticJournal.time)).first()

    def after(self, s, time=None):
        if not time:
            time = self.time_added(s)
        return self._base_sibling_query(s, KIT_ADDED).filter(StatisticJournal.time > time). \
            order_by(StatisticJournal.time).first()

    def undo(self, s):
        time = self.time_added(s)
        s.delete(self)
        before = self.before(s, time)
        if before:
            before._remove_statistic(s, KIT_RETIRED)
            after = self.after(s, time)
            if after:
                before._add_timestamp(s, KIT_RETIRED, after.time_added(s))

    def time_range(self, s):
        return None, None

    def _add_children(self, s, model, depth=0, statistics=None, time=None, own_models=True):
        pass

    def __str__(self):
        return f'KitModel "{self.name}"'




from sqlalchemy import Column, Text, Integer, ForeignKey
from sqlalchemy.orm import relationship, backref

from ..support import Base
from ..types import Time, ShortCls
from ..utils import add
from ...lib import to_time


class FileHash(Base):
    '''
    This exists so that hash can be used to connect entities (eg activity topics).
    Entries are never deleted from here (but it's created empty on database updates).
    '''

    __tablename__ = 'file_hash'

    id = Column(Integer, primary_key=True)
    hash = Column(Text, nullable=False, index=True)

    @classmethod
    def get_or_add(cls, s, hash):
        instance = s.query(FileHash).filter(FileHash.hash == hash).one_or_none()
        if not instance:
            instance = add(s, FileHash(hash=hash))
        return instance


class FileScan(Base):

    __tablename__ = 'file_scan'

    path = Column(Text, nullable=False, primary_key=True)
    owner = Column(ShortCls, nullable=False, primary_key=True)
    last_scan = Column(Time, nullable=False)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('file_scan', cascade='all, delete-orphan',
                                                         passive_deletes=True, uselist=False))

    @classmethod
    def add(cls, s, path, owner, hash):
        return add(s, FileScan(path=path, owner=owner, last_scan=to_time(0.0), file_hash=FileHash.get_or_add(s, hash)))

    def __str__(self):
        return self.path




from bokeh.io import output_file, show
from bokeh.plotting import figure
from math import log10

from ch2.data import *
from ch2.data.plot.utils import evenly_spaced_hues
from ch2.data.response import *
from ch2.lib.utils import group_to_dict
from ch2.sql import *
from ch2.stats.read.segment import SegmentReader
from ch2.jupyter.decorator import template


@template
def fit_ff_segments(group, *segment_names):

    f'''
    # Fit FF Parameters for {group} to {', '.join(segment_names)}

    This notebook allows you to estimate a personal time scale (decay period) for the
    [FF model](https://andrewcooke.github.io/choochoo/impulse) using your times (more exactly, the speed)
    on the named segments.

    The resulting value can be used in the configuration so that the Fitness parameter more accurately
    reflects your personal rate of adaption.

    For more information see the [documentation](https://andrewcooke.github.io/choochoo/ff-fitting).
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.  
    We reduce the HR data to hourly values so that we have smaller arrays (for faster processing).
    '''

    s = session('-v2')

    hr10 = statistics(s, HR_IMPULSE_10, constraint=ActivityGroup.from_name(s, 'all'))
    print(hr10.describe())
    segments = [s.query(Segment).filter(Segment.name == segment_name).one() for segment_name in segment_names]
    for segment in segments:
        print(segment.name, segment.distance)
    kit_statistic = StatisticName.from_name(s, 'kit', SegmentReader, ActivityGroup.from_name(s, group))
    journals_by_kit_by_segment = \
        {segment: group_to_dict(s.query(StatisticJournalText.value, SegmentJournal).
                                join(ActivityJournal, SegmentJournal.activity_journal_id == ActivityJournal.id).
                                join(StatisticJournalText, StatisticJournalText.source_id == ActivityJournal.id).
                                filter(StatisticJournalText.statistic_name_id == kit_statistic.id,
                                       SegmentJournal.segment == segment).all())
         for segment in segments}
    times_by_kit_by_segment = \
        {segment: {kit: drop_empty(statistics(s, SEGMENT_TIME, sources=journals)).dropna()
                   for kit, journals in journals_by_kit_by_segment[segment].items()}
         for segment in segments}
    performances = []
    for segment in segments:
        for kit, times in times_by_kit_by_segment[segment].items():
            if len(times.columns) and len(times.index) > 2:
                times.index = times.index.round('1H')
                performances.append(Series(segment.distance / times.iloc[:, 0], times.index,
                                           name=f'{segment.name}/{kit}'))
    n_performances = sum(len(performance) for performance in performances)

    hr3600 = sum_to_hour(hr10, HR_IMPULSE_10)

    def copy_of_performances():
        return [performance.copy() for performance in performances]

    for performance in performances:
        print(performance)

    '''
    ## Define Plot Routine

    Below the `params` are (currently) log10_period and log10_start - the time period for decay and the initial
    response (fitness) value.  We use log10 so that the values cannot be negative.
    '''

    output_file(filename='/dev/null')

    def plot_params(params, rejected=tuple()):
        response = calc_response(hr3600, params)
        predicted = calc_observed(restrict_response(response, performances), performances)
        f = figure(title=fmt_params(params), plot_width=500, plot_height=450, x_axis_type='datetime')
        f.line(x=response.index, y=response, color='grey')
        for color, prediction, performance in zip(evenly_spaced_hues(len(performances)), predicted, performances):
            f.circle(x=prediction.index, y=prediction, color=color, legend_label=performance.name)
        for i, t in rejected:
            f.x(x=t, y=predicted[i].loc[t], color='black', size=10)
        f.legend.location = 'bottom_left'
        show(f)

    '''
    ## Plot Initial Response
    
    Trim and resample data; evaluate and plot our initial model.
    '''

    initial_period = log10(42 * 24)
    plot_params((initial_period,))

    '''
    ## Explore Effect of Start
    '''
    plot_params((initial_period, log10(1000)))
    plot_params((initial_period, log10(3000)))

    '''
    ## Fit Model using L1
    
    Adjust tol according to the 'fun' value in the result (tol is the tolerance in that value).
    
    Note how the period (printed, in hours) varies as points are rejected.
    '''

    result, rejected = fit_ff_params(hr3600, (initial_period,), copy_of_performances(),
                                     method='L1', tol=0.01,
                                     max_reject=n_performances // 2, threshold=(2, 0.2))
    print(result)
    plot_params(result.x, rejected=rejected)

    result, rejected = fit_ff_params(hr3600, (initial_period, 0), copy_of_performances(),
                                     method='L1', tol=0.01,
                                     max_reject=n_performances // 2, threshold=(2, 0.2))
    print(result)
    plot_params(result.x, rejected=rejected)

    '''
    ## Fit Model using L2
    
    Using different methods gives us some idea of how susceptible the value is to processing assumptions.
    
    Note that the thresholds change for the L2 norm.
    '''

    initial_period = log10(42 * 24)
    result, rejected = fit_ff_params(hr3600, (initial_period,), copy_of_performances(),
                                     method='L2', tol=0.01,
                                     max_reject=n_performances // 2, threshold=(500, 50))
    print(result)
    plot_params(result.x, rejected=rejected)

    result, rejected = fit_ff_params(hr3600, (initial_period, 0), copy_of_performances(),
                                     method='L2', tol=0.01,
                                     max_reject=n_performances // 2, threshold=(500, 50))
    print(result)
    plot_params(result.x, rejected=rejected)




import datetime as dt
from json import dumps
from logging import getLogger

from pendulum.tz import get_local_timezone
from sqlalchemy import Column, Integer, Text, ForeignKey, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship, backref

from .source import SourceType, Source, Interval
from .statistic import StatisticJournal, STATISTIC_JOURNAL_CLASSES
from .system import SystemConstant
from ..support import Base
from ..types import Date, Json, Sched, Sort
from ..utils import add
from ...lib.date import local_date_to_time
from ...lib.schedule import Schedule

log = getLogger(__name__)


class Topic:
    '''
    A topic groups together a set of fields.  At it's simplest, think of it as a title in the diary.
    Topics can contain child topics, giving a tree-like structure, but relationships must be defined in
    concrete classes.
    '''

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='0')


class DiaryTopic(Base, Topic):
    '''
    DiaryTopics are associated with a schedule, which means that may only be displayed on certain dates.
    '''

    __tablename__ = 'diary_topic'

    parent_id = Column(Integer, ForeignKey('diary_topic.id'), nullable=True)
    schedule = Column(Sched, nullable=False)
    start = Column(Date)
    finish = Column(Date)

    @declared_attr
    def children(cls):
        # http://docs.sqlalchemy.org/en/latest/orm/self_referential.html
        return relationship('DiaryTopic', backref=backref('parent', remote_side=[cls.id]))

    def __init__(self, id=None, parent=None, parent_id=None, schedule=None, name=None, description=None, sort=None):
        # Topic instances are only created in config.  so we intercept here to
        # duplicate data for start and finish - it's not needed elsewhere.
        if not isinstance(schedule, Schedule):
            schedule = Schedule(schedule)
        self.id = id
        self.parent = parent
        self.parent_id = parent_id
        self.schedule = schedule
        self.name = name
        self.description = description
        self.sort = sort
        self.start = schedule.start
        self.finish = schedule.finish

    def __str__(self):
        return 'DiaryTopic "%s" (%s)' % (self.name, self.schedule)


class ActivityTopic(Base, Topic):

    __tablename__ = 'activity_topic'

    parent_id = Column(Integer, ForeignKey('activity_topic.id'), nullable=True)
    activity_group_id = Column(Integer, ForeignKey('activity_group.id'), nullable=True)
    activity_group = relationship('ActivityGroup')

    @declared_attr
    def children(cls):
        # http://docs.sqlalchemy.org/en/latest/orm/self_referential.html
        return relationship('ActivityTopic', backref=backref('parent', remote_side=[cls.id]))

    def __str__(self):
        return 'ActivityTopic "%s" (%s)' % (self.name, self.activity_group)


class TopicField:

    id = Column(Integer, primary_key=True)
    sort = Column(Sort, nullable=False, server_default='0')
    model = Column(Json, nullable=False, server_default=dumps({}))

    @declared_attr
    def statistic_name_id(cls):
        # https://stackoverflow.com/questions/24666261/sqlalchemy-mixins-foreignkeys-and-declared-attr
        return Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)

    @declared_attr
    def statistic_name(cls):
        return relationship('StatisticName')


class DiaryTopicField(Base, TopicField):

    # diary topic fields are associated with statistics whose constraints are the diary topic.
    # this lets us use the same name under different 'headings'.

    __tablename__ = 'diary_topic_field'

    diary_topic_id = Column(Integer, ForeignKey('diary_topic.id', ondelete='cascade'), nullable=False)
    diary_topic = relationship('DiaryTopic',
                               backref=backref('fields', cascade='all, delete-orphan',
                                               passive_deletes=True,
                                               order_by='DiaryTopicField.sort'))
    schedule = Column(Sched, nullable=False, server_default='')
    UniqueConstraint('statistic_name_id')

    def __str__(self):
        return 'DiaryTopicField "%s"/"%s"' % (self.diary_topic.name, self.statistic_name.name)


class ActivityTopicField(Base, TopicField):

    # activity topic fields are associated with statistics whose constraints are the activity group.
    # this correlates to some extent with the activity topic, which is also specific to an activity group
    # (so you will never see a field for group X if it has parent activity for group Y).

    NAME = 'Name'

    __tablename__ = 'activity_topic_field'

    # unlike diary topic fields, this can have a null parent.  that means that it is a child to the
    # 'top' of the activity.
    # the 'Name' field with a null parent is the activity title,
    # this is loaded with a default (based on file name) when data are read from FIT files.
    activity_topic_id = Column(Integer, ForeignKey('activity_topic.id', ondelete='cascade'), nullable=True)
    activity_topic = relationship('ActivityTopic',
                                  backref=backref('fields', cascade='all, delete-orphan',
                                                  passive_deletes=True,
                                                  order_by='ActivityTopicField.sort'))
    UniqueConstraint('statistic_name_id')

    def __str__(self):
        return 'ActivityTopicField "%s"/"%s"' % (self.activity_topic.name, self.statistic_name.name)


class Cache:

    def __init__(self, s, source, time, query):
        self.__session = s
        self.__source = source
        self.__time = time
        self.__cache = {field_id: statistic for field_id, statistic in query.all()}

    def __getitem__(self, field):
        if field.id in self.__cache:
            return self.__cache[field.id]
        else:
            name = field.statistic_name
            return add(self.__session,
                       STATISTIC_JOURNAL_CLASSES[name.statistic_journal_type](
                           statistic_name=name, source=self.__source, time=self.__time))

    def __len__(self):
        return len(self.__cache)


class DiaryTopicJournal(Source):

    __tablename__ = 'diary_topic_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    date = Column(Date, nullable=False, index=True)
    UniqueConstraint(date)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.DIARY_TOPIC
    }

    @classmethod
    def get_or_add(cls, s, date):
        instance = s.query(DiaryTopicJournal).filter(DiaryTopicJournal.date == date).one_or_none()
        if not instance:
            instance = add(s, DiaryTopicJournal(date=date))
        return instance

    def cache(self, s):
        return Cache(s, self, local_date_to_time(self.date),
                     s.query(DiaryTopicField.id, StatisticJournal).
                     filter(DiaryTopicField.statistic_name_id == StatisticJournal.statistic_name_id,
                            StatisticJournal.source_id == self.id))

    def __str__(self):
        return 'DiaryTopicJournal from %s' % self.date

    @classmethod
    def check_tz(cls, system, s):
        tz = get_local_timezone()
        db_tz = system.get_constant(SystemConstant.TIMEZONE, none=True)
        if db_tz is None:
            db_tz = system.set_constant(SystemConstant.TIMEZONE, '')
        if db_tz != tz.name:
            cls.__reset_timezone(s)
            system.set_constant(SystemConstant.TIMEZONE, tz.name, force=True)

    @classmethod
    def __reset_timezone(cls, s):
        log.warning('Timezone has changed')
        log.warning('Recalculating times for TopicJournal entries')
        for tj in s.query(DiaryTopicJournal).all():
            tj.time = local_date_to_time(tj.date)
        Interval.dirty_all(s)

    def time_range(self, s):
        start = local_date_to_time(self.date)
        return start, start + dt.timedelta(days=1)


class ActivityTopicJournal(Source):

    __tablename__ = 'activity_topic_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('activity_topic_journal', uselist=False))
    UniqueConstraint(file_hash_id)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY_TOPIC
    }

    @classmethod
    def get_or_add(cls, s, file_hash):
        instance = s.query(ActivityTopicJournal).filter(ActivityTopicJournal.file_hash == file_hash).one_or_none()
        if not instance:
            instance = add(s, ActivityTopicJournal(file_hash=file_hash))
        return instance

    def cache(self, s):
        return Cache(s, self, self.file_hash.activity_journal.start,
                     s.query(ActivityTopicField.id, StatisticJournal).
                     filter(ActivityTopicField.statistic_name_id == StatisticJournal.statistic_name_id,
                            StatisticJournal.source_id == self.id))




import itertools as it
from collections import namedtuple, OrderedDict

# todo - remove 's' units from timestamps if they are datetime instances
from ...lib.data import dict_to_attr


def no_filter(data):
    return data


def no_bad_values(data):
    for name, (values, units) in data:
        if values is not None:
            yield name, (values, units)


def no_unknown_fields(data):
    for name, values_or_pair in data:
        if name[0].islower():
            yield name, values_or_pair


def no_names(data):
    for name, values_or_pair in data:
        yield values_or_pair


def no_values(data):
    for name, values_or_pair in data:
        yield name


def no_units(data):
    for name, (values, units) in data:
        if values is not None:
            yield name, values


def to_hex(data):
    for name, (values, units) in data:
        if values is None:
            yield name, (values, units)
        else:
            if len(values) > 1:  # single values are best displayed as integers (common data type)
                try:
                    values = ('0x'+values.hex(),)
                except AttributeError:
                    pass
            yield name, (values, units)


def append_units(data, separator=''):
    for name, (values, units) in data:
        if values is None:  # preserve bad values as bad
            yield name, None
        elif units:
            yield name, tuple(str(value) + separator + units for value in values)
        else:
            yield name, tuple(str(value) for value in values)


def join_values(data, separator=','):
    for name, values in data:
        if values is None:
            yield name, values
        else:
            yield name, separator.join(values)


def fix_degrees(data, new_units='°'):
    for name, (values, units) in data:
        if values is not None and units == 'semicircles':
            values = tuple(value * 180 / 2**31 for value in values)
            units = new_units
        yield name, (values, units)


def unpack_single_bytes(data):
    for name, (values, units) in data:
        if values is None or not (isinstance(values, bytes) and len(values) == 1):
            yield name, (values, units)
        else:
            yield name, (values[0], units)


def unique_names(data):
    known = set()
    for name, values_or_pair in data:
        if name not in known:
            yield name, values_or_pair
        known.add(name)


# this is intended for slurping to a dict - it does not (cannot) remove earlier entries,
# but accumulates duplicates in later entries.
def merge_duplicates(data):
    known = {}
    data = list(data)
    for name, (values, units) in data:
        if name in known and known[name][0] is not None:
            values = known[name][0] + values
        known[name] = (values, units)
        yield name, known[name]


def restrict_names(names):
    def filter(data):
        for name, values_or_pair in data:
            if name in names:
                yield name, values_or_pair
    return filter


def chain(*filters):
    def expand(data, filters=filters):
        if filters:
            filter, filters = filters[0], filters[1:]
            return filter(expand(data, filters=filters))
        else:
            return data
    return expand


class Record(namedtuple('BaseRecord', 'name, number, identity, timestamp, data')):

    __slots__ = ()

    def is_known(self):
        return self.name[0].islower()

    def data_with(self, **kargs):
        return it.chain(self.data, kargs.items())

    def into(self, container, *filters, _cls=None, **extras):
        if not _cls:
            _cls = Record
        return _cls(self.name, self.number, self.identity, self.timestamp,
                    container(chain(*filters)(self.data_with(**extras))))

    def as_dict(self, *filters, **extras):
        return self.into(OrderedDict, *filters, **extras, _cls=DictRecord)

    def as_names(self, *filters, **extras):
        return self.into(tuple, *(no_values,)+filters, **extras)

    def as_values(self, *filters, **extras):
        return self.into(tuple, *(no_names,)+filters, **extras)

    def force(self, *filters, **extras):
        if filters or extras:
            return self.into(type(self.data), *filters, **extras)
        else:
            return self


class LazyRecord(Record):

    __slots__ = ()

    def force(self, *filters, **extras):
        return self.as_dict(*filters, **extras)


class Values:

    def __init__(self, attr):
        self.__attr = attr

    def __getattr__(self, name):
        (values, units) = getattr(self.__attr, name)
        return values[0]


class DictRecord(Record):

    def __new__(cls, *args, **kargs):
        self = super().__new__(cls, *args, **kargs)
        self.attr = dict_to_attr(self.data)
        self.value = Values(self.attr)
        return self

    def data_with(self, **kargs):
        return it.chain(self.data.items(), kargs.items())





import datetime as dt
from abc import abstractmethod
from logging import getLogger

from ...lib.date import local_date_to_time, to_date
from ...lib.log import log_current_exception
from ...sql import ActivityGroup, ActivityJournal
from ...sql.tables.pipeline import Pipeline, PipelineType
from ...stats.pipeline import BasePipeline

log = getLogger(__name__)


def read_pipeline(session, date, schedule=None):
    '''
    schedule only sent for summary views.
    '''
    date = to_date(date)   # why is this needed?
    for pipeline in Pipeline.all(session, PipelineType.DIARY):
        log.info(f'Building {pipeline.cls} ({pipeline.args}, {pipeline.kargs})')
        instance = pipeline.cls(*pipeline.args, **pipeline.kargs)
        if isinstance(instance, Reader):
            data = list(instance.read(session, date, schedule=schedule))
            if data:
                yield data


class Reader(BasePipeline):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)

    def read(self, s, date, schedule=None):
        try:
            if schedule:
                yield from self._read_schedule(s, date, schedule)
            else:
                yield from self._read_date(s, date)
        except Exception as e:
            log_current_exception(e)

    @abstractmethod
    def _read_schedule(self, s, date, schedule):
        raise NotImplementedError(self.__class__)
        yield

    @abstractmethod
    def _read_date(self, s, date):
        raise NotImplementedError(self.__class__)
        yield


class JournalDiary(Reader):

    def _read_date(self, s, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for ajournal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield from self._read_journal_date(s, ajournal, date)

    @abstractmethod
    def _read_journal_date(self, s, ajournal, date):
        raise NotImplementedError(self.__class__)
        yield




from logging import getLogger
from os import getpid
from time import time, sleep

import psutil as ps
from sqlalchemy import Column, Text, Integer, Float, UniqueConstraint

from ..support import SystemBase
from ..types import Time, ShortCls
from ..utils import add
from ...lib import now

log = getLogger(__name__)


class SystemConstant(SystemBase):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)

    @classmethod
    def get(cls, s, name, none=False):
        instance = s.query(SystemConstant).filter(SystemConstant.name == name).one_or_none()
        if instance is None:
            if not none:
                raise Exception(f'No value for {name}')
            else:
                log.debug(f'Read {name}=None')
                return None
        else:
            log.debug(f'Read {name}={instance.value}')
            return instance.value

    @classmethod
    def set(cls, s, name, value, force=False):
        if force:
            cls.delete(s, name)
        s.add(SystemConstant(name=name, value=value))
        s.commit()
        log.debug(f'Set {name}={value}')
        return value

    @classmethod
    def delete(cls, s, name):
        s.query(SystemConstant).filter(SystemConstant.name == name).delete()
        s.commit()

    TIMEZONE = 'timezone'
    JUPYTER_URL = 'jupyter-url'
    WEB_URL = 'web-url'
    LAST_GARMIN = 'last-garmin'
    DB_VERSION = 'db-version'


class Process(SystemBase):

    __tablename__ = 'process'

    id = Column(Integer, primary_key=True)
    owner = Column(ShortCls, nullable=False, index=True)
    pid = Column(Integer, nullable=False, index=True)
    start = Column(Time, nullable=False, default=time)
    command = Column(Text, nullable=True)
    log = Column(Text, nullable=True)

    @classmethod
    def run(cls, s, owner, cmd, log_name):
        process = ps.Popen(args=cmd, shell=True)
        log.debug(f'Adding command [{cmd}] (PID {process.pid})')
        s.add(Process(command=cmd, owner=owner, pid=process.pid, log=log_name))
        s.commit()
        return process

    @classmethod
    def delete(cls, s, owner, pid, delta_seconds=3):
        process = s.query(Process).filter(Process.owner == owner, Process.pid == pid).one()
        process.__kill(delta_seconds=delta_seconds)
        log.debug(f'Deleting record for process {process.pid}')
        s.delete(process)
        s.commit()

    @classmethod
    def delete_all(cls, s, owner, delta_seconds=3):
        for process in s.query(Process).filter(Process.owner == owner).all():
            if process.pid == getpid():
                log.debug(f'Not killing self (PID {process.pid})')
            else:
                process.__kill(delta_seconds=delta_seconds)
                log.debug(f'Deleting record for process {process.pid}')
                s.delete(process)
        s.commit()

    @classmethod
    def exists_any(cls, s, owner):
        for process in s.query(Process).filter(Process.owner == owner).all():
            if process.__still_running():
                return True
        cls.delete_all(s, owner)
        return False

    def __still_running(self, delta_seconds=3):
        return exists(self.pid, self.start, delta_seconds=delta_seconds)

    def __kill(self, delta_seconds=3):
        if self.__still_running(delta_seconds=delta_seconds):
            log.debug(f'Killing process {self.pid}')
            ps.Process(self.pid).kill()


# if we're given a delta time, the start time of the process and database records have to match
# (or it's some other unrelated process because the process counter wrapped round)
# if they don't match we return false because the original process is not running.
def exists(pid, time, delta_seconds=3, zombie_seconds=10):
    if not ps.pid_exists(pid):
        log.debug(f'Process {pid} does not exist')
        return False
    process = ps.Process(pid)
    if process.status() == ps.STATUS_ZOMBIE:
        try:
            log.warning(f'Waiting for zombie {pid}')
            process.wait(zombie_seconds)
        except ps.TimeoutExpired:
            log.debug(f'Timeout waiting for zombie {pid}')
        return False
    elif delta_seconds:
        creation_ok = abs(process.create_time() - time.timestamp()) < delta_seconds
        if creation_ok:
            log.debug(f'Process {pid} still exists')
        else:
            log.debug(f'Creation time for process {pid} incorrect')
        return creation_ok
    else:
        log.debug(f'Assuming process {pid} is same as expected')
        return True   # if delta_seconds = 0 don't check, just assume same


class Progress(SystemBase):

    __tablename__ = 'progress'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, unique=True)
    pid = Column(Integer, nullable=False, index=True)
    start = Column(Time, nullable=False)
    percent = Column(Integer, nullable=False, default=0)
    error = Column(Text, nullable=True)

    def __str__(self):
        return f'Progress {self.name} / {self.pid}'

    @classmethod
    def create(cls, s, name, delta_seconds=3):
        progress = s.query(Progress).filter(Progress.name == name).one_or_none()
        if progress:
            if progress.percent != 100 and exists(progress.pid, progress.start, delta_seconds=delta_seconds):
                raise Exception(f'Progress {name} already exists (PID {progress.pid})')
            else:
                log.debug(f'Removing old progress {name} / {progress.pid}')
                s.delete(progress)
                s.commit()
        us = ps.Process(getpid())
        add(s, Progress(name=name, pid=us.pid, start=us.create_time()))

    @classmethod
    def update(cls, s, name, **kargs):
        progress = s.query(Progress).filter(Progress.name == name, Progress.pid == getpid()).one_or_none()
        if not progress:
            raise Exception(f'No existing progress for {name} / {getpid()}')
        for name in kargs:
            value = kargs[name]
            log.debug(f'Setting progress {name}={value}')
            if hasattr(progress, name):
                setattr(progress, name, value)
            else:
                raise AttributeError(name)

    @classmethod
    def get_percent(cls, s, name, delta_seconds=3):
        progress = s.query(Progress).filter(Progress.name == name).one_or_none()
        if progress is None or not exists(progress.pid, progress.start, delta_seconds=delta_seconds):
            log.debug(f'No percent for {name}')
            return None
        else:
            log.debug(f'Progress for {name} is {progress.percent}%')
            return progress.percent

    @classmethod
    def wait_for_progress(cls, s, name, timeout=60, delta_seconds=3, pause=1):

        def pid(progress):
            if progress:
                pid = progress.pid
                s.expire(progress)
                return pid

        log.debug(f'Waiting for progress {name}')
        start = now()
        initial = s.query(Progress).filter(Progress.name == name).one_or_none()
        log.debug(f'Initial progress: {initial}')
        if initial and exists(initial.pid, initial.start):
            return

        # use PIDs so that we don't need to worry about expired objects
        initial = pid(initial)
        while True:
            sleep(pause)
            current = pid(s.query(Progress).filter(Progress.name == name).one_or_none())
            log.debug(f'Waiting for PID - comparing {initial} and {current}')
            if current != initial:
                return
            elif (now() - start).total_seconds() >= timeout:
                raise Exception(f'Did not find progress {name} before {timeout}s')




from inspect import getfullargspec
from logging import getLogger
from pkgutil import iter_modules


log = getLogger(__name__)


def read_package(package):
    '''
    This reads files (modules) from a package and extracts functions with the same name as the file.

    That sounds weird, but it's what we want for config and jupyter - see the config.profile and jupyter.template
    packages.

    Note that the file (module) must be imported into __init__ to be visible (more exactly, it must have been
    imported into Python somehow previously).
    '''
    log.debug(f'Reading {package.__file__}')
    log.debug(dir(package))
    for importer, modname, ispkg in iter_modules(package.__path__):
        try:
            module = getattr(package, modname)
            function = getattr(module, modname)
            if hasattr(function, '_original'):
                # drop through template decorator
                argspec = getfullargspec(function._original)
            else:
                argspec = getfullargspec(function)
            yield modname, (function, argspec)
        except AttributeError:
            log.debug(f'Skipping {modname}')
            log.debug('(if this is unexpected, check that the file is imported at the package level)')




from .data import sorted_numeric_labels, left_interpolate, bookend, median
from .date import local_date_to_time, to_date, time_to_local_time, format_date, format_seconds, format_time, \
    local_time_to_time, time_to_local_date, to_time, datetime_to_epoch, local_time_or_now, now, \
    is_local_time
from .utils import short_str, sign, format_metres, format_percent, format_watts, groupby_tuple, drop_trailing_slash, \
    format_km




# lets us define a template in python that IntelliJ will validate without flagging an unused code error.
TEMPLATE = lambda *args: False

# need to import these so they are available for introspection
from . import activity_details, all_activities, calendar, health, similar_activities, some_activities, \
    fit_power_parameters, define_segment, month, heart_rate, fit_ff_segments, route_sentinel_3d, \
    nearby_activities, gmap_activities



from logging import getLogger

from ...commands.kit import finish, change, start
from ...lib import local_date_to_time, now
from ...sql import KitGroup, KitComponent, KitItem
from ...sql.tables.kit import MODELS, ITEMS, ITEM, COMPONENT, MODEL, INDIVIDUAL, POPULATION, GROUP

log = getLogger(__name__)


class Kit:
    '''
    These are / should be very thin wrappers around the equivalent kit commands.
    We don't try to implement all commands, just the basics.
    If necessary, the user can drop down to the command line.
    '''

    @staticmethod
    def _delete_item_models(groups):
        for group in groups:
            for item in group[ITEMS]:
                del item[MODELS]
        return groups

    @staticmethod
    def read_snapshot(request, s, date):
        return [group.to_model(s, depth=3, statistics=INDIVIDUAL, time=local_date_to_time(date))
                for group in s.query(KitGroup).order_by(KitGroup.name).all()]

    @staticmethod
    def read_edit(request, s):
        return [group.to_model(s, depth=3, statistics=None, time=now(), own_models=False)
                for group in s.query(KitGroup).order_by(KitGroup.name).all()]

    @staticmethod
    def read_statistics(request, s):
        return [component.to_model(s, depth=3, statistics=POPULATION)
                for component in s.query(KitComponent).order_by(KitComponent.name).all()]

    @staticmethod
    def read_items(request, s):
        return [item.to_model(s, depth=0)
                for item in s.query(KitItem).order_by(KitItem.name).all()]

    @staticmethod
    def write_retire_item(request, s):
        data = request.json
        log.debug(data)
        finish(s, data[ITEM], None, True)

    @staticmethod
    def write_replace_model(request, s):
        data = request.json
        log.debug(data)
        change(s, data[ITEM], data[COMPONENT], data[MODEL], None, False, False)

    @staticmethod
    def write_add_component(request, s):
        data = request.json
        log.debug(data)
        change(s, data[ITEM], data[COMPONENT], data[MODEL], None, True, False)

    @staticmethod
    def write_add_group(request, s):
        data = request.json
        log.debug(data)
        start(s, data[GROUP], data[ITEM], None, True)




import datetime as dt

from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint, desc
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time, Sort, ShortCls, NullStr
from ...lib.date import format_time, local_date_to_time, local_time_to_time


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    def __str__(self):
        return 'ActivityGroup "%s"' % self.name

    @classmethod
    def from_name(cls, s, name, optional=False):
        if optional:
            return s.query(ActivityGroup).filter(ActivityGroup.name.ilike(name)).one_or_none()
        else:
            return s.query(ActivityGroup).filter(ActivityGroup.name.ilike(name)).one()


class ActivityJournal(Source):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    activity_group_id = Column(Integer, ForeignKey('activity_group.id'), nullable=False)
    activity_group = relationship('ActivityGroup')
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('activity_journal', uselist=False))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_group_id, start)
    UniqueConstraint(file_hash_id)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish

    @classmethod
    def at_date(cls, s, date):
        day = local_date_to_time(date)
        return s.query(ActivityJournal).filter(ActivityJournal.start >= day,
                                               ActivityJournal.start < day + dt.timedelta(days=1)).all()

    @classmethod
    def at_local_time(cls, s, local_time):
        time = local_time_to_time(local_time)
        return s.query(ActivityJournal).filter(ActivityJournal.start == time).one()

    @classmethod
    def from_id(cls, s, id):
        return s.query(ActivityJournal).filter(ActivityJournal.id == id).one()

    @classmethod
    def before_local_time(cls, s, local_time):
        time = local_date_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start < time). \
            order_by(desc(ActivityJournal.start)). \
            limit(1).one_or_none()

    @classmethod
    def after_local_time(cls, s, local_time):
        time = local_date_to_time(local_time)
        return s.query(ActivityJournal). \
            filter(ActivityJournal.start >= (time + dt.timedelta(days=1))). \
            order_by(ActivityJournal.start). \
            limit(1).one_or_none()

    @classmethod
    def number_of_activities(cls, s):
        return s.query(ActivityJournal).count()


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))


class ActivityBookmark(Base):

    __tablename__ = 'activity_bookmark'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), nullable=False)
    activity_journal = relationship('ActivityJournal')
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    constraint = Column(NullStr, index=True)
    UniqueConstraint(activity_journal_id, start, finish, owner, constraint)

    def __str__(self):
        return 'ActivityBookmark from %s - %s' % (format_time(self.start), format_time(self.finish))





import datetime as dt
from logging import getLogger
from re import sub

from . import Reader
from ..calculate.segment import SegmentCalculator
from ..names import SEGMENT_TIME, SEGMENT_HEART_RATE
from ...diary.database import summary_column
from ...diary.model import value, text, optional_text
from ...lib.date import local_date_to_time
from ...sql.tables.segment import SegmentJournal, Segment
from ...sql.tables.statistic import StatisticJournal, StatisticName

log = getLogger(__name__)


def segments_for_activity(s, ajournal):
    return s.query(SegmentJournal). \
        filter(SegmentJournal.activity_journal == ajournal). \
        order_by(SegmentJournal.start).all()


class SegmentDiary(Reader):

    @optional_text('Segments')
    def _read_date(self, s, date):
        tomorrow = local_date_to_time(date + dt.timedelta(days=1))
        today = local_date_to_time(date)
        for sjournal in s.query(SegmentJournal).join(Segment). \
                filter(SegmentJournal.start >= today,
                       SegmentJournal.start < tomorrow). \
                order_by(SegmentJournal.start).all():
            stats = [value(sub('^Segment ', '', field.statistic_name.name), field.value,
                           units=field.statistic_name.units, measures=field.measures_as_model(date))
                     for field in (self.__field(s, date, sjournal, name)
                                   for name in (SEGMENT_TIME, SEGMENT_HEART_RATE))
                     if field]
            if stats:
                yield [text(sjournal.segment.name, tag='segment')] + stats

    @staticmethod
    def __field(s, date, sjournal, name):
        return StatisticJournal.at_date(s, date, name, SegmentCalculator, sjournal.segment, source_id=sjournal.id)

    @optional_text('Segments')
    def _read_schedule(self, s, date, schedule):
        for segment in s.query(Segment).all():
            segment_rows = [list(summary_column(s, schedule, date, name))
                            for name in self.__names(s, segment, SEGMENT_TIME, SEGMENT_HEART_RATE)]
            segment_rows = list(filter(bool, segment_rows))
            if segment_rows:
                yield [text(segment.name)] + segment_rows

    @staticmethod
    def __names(s, segment, *names):
        for name in names:
            sname = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.constraint == segment,
                       StatisticName.owner == SegmentCalculator).one_or_none()
            if sname:
                yield sname




import datetime as dt
from json import loads

from sqlalchemy import asc, desc

from . import Reader
from ..calculate.response import Response, ResponseCalculator
from ...diary.model import text, optional_text, value
from ...lib.date import local_date_to_time, to_time
from ...lib.schedule import Schedule
from ...sql.tables.constant import Constant
from ...sql.tables.statistic import StatisticJournal, StatisticName, TYPE_TO_JOURNAL_CLASS


class ResponseDiary(Reader):

    def __init__(self, *args, fitness=None, fatigue=None, **kargs):
        self.fitness = self._assert('fitness', fitness)
        self.fatigue = self._assert('fatigue', fatigue)
        super().__init__(*args, **kargs)

    def _read_date(self, s, date):
        yield from self._read_schedule(s, date, Schedule('d'))

    @optional_text('SHRIMP')
    def _read_schedule(self, s, date, schedule):
        for response in self.fitness + self.fatigue:
            yield from self._read_single(s, date, schedule, response, schedule.frame_type == 'd')

    def _read_single(self, s, date, schedule, constant_name, display_range, ranges=('all', '90d', '30d')):
        start_time = local_date_to_time(schedule.start_of_frame(date))
        finish_time = local_date_to_time(schedule.next_frame(date))
        response = Response(**loads(Constant.get(s, constant_name).at(s, start_time).value))
        start = self._read(s, response.dest_name, start_time, finish_time, asc)
        finish = self._read(s, response.dest_name, start_time, finish_time, desc)
        if start and finish and start.value != finish.value:
            model = [text(response.dest_name),
                     value('From', int(start.value)), value('To', int(finish.value)),
                     text('⇧' if start.value < finish.value else '⇩')]
            if display_range:
                for range in ranges:
                    limits = self._range(s, response.dest_name, start, finish_time,
                                         None if range == 'all' else dt.timedelta(days=int(range[:-1])))
                    model.append([text(f'Over {range}', tag=range),
                                  value('Lo', int(limits[0].value)), value('Hi', int(limits[1].value))])
            yield model

    @staticmethod
    def _read(s, name, start_time, finish_time, direcn):
        return s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == ResponseCalculator,
                   StatisticJournal.time >= start_time,
                   StatisticJournal.time < finish_time). \
            order_by(direcn(StatisticJournal.time)). \
            limit(1).one_or_none()

    @staticmethod
    def _range(s, name, value, finish_time, period):
        jtype = TYPE_TO_JOURNAL_CLASS[type(value.value)]
        start_time = finish_time - period if period else to_time(0.0)
        q = s.query(jtype). \
            join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == ResponseCalculator,  # todo - owner_in
                   jtype.time >= start_time,
                   jtype.time < finish_time)
        return (q.order_by(asc(jtype.value)).limit(1).one_or_none(),
                q.order_by(desc(jtype.value)).limit(1).one_or_none())




from sqlalchemy import Column, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..types import Time
from ...lib.date import format_time


class MonitorJournal(Source):

    __tablename__ = 'monitor_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('monitor_journal', uselist=False))
    start = Column(Time, nullable=False, index=True)
    finish = Column(Time, nullable=False, index=True)
    UniqueConstraint(file_hash_id)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.MONITOR
    }

    def __str__(self):
        return 'Monitor Journal %s to %s' % (format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish





import datetime as dt
from json import loads
from logging import getLogger

from . import ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator
from ..names import ELEVATION, DISTANCE, M, POWER_ESTIMATE, HEART_RATE, ACTIVE_DISTANCE, MSR, SUM, CNT, MAX, \
    summaries, ACTIVE_SPEED, ACTIVE_TIME, AVG, S, KMH, MIN_KM_TIME_ANY, MIN, MED_KM_TIME_ANY, PERCENT_IN_Z_ANY, PC, \
    TIME_IN_Z_ANY, MAX_MED_HR_M_ANY, W, BPM, MAX_MEAN_PE_M_ANY, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_TIME, \
    CLIMB_GRADIENT, TOTAL_CLIMB, HR_ZONE, TIME, like, MEAN_POWER_ESTIMATE, ENERGY_ESTIMATE, SPHERICAL_MERCATOR_X, \
    SPHERICAL_MERCATOR_Y, DIRECTION, DEG, ASPECT_RATIO, FITNESS_D_ANY, FATIGUE_D_ANY, _delta, FF, CLIMB_POWER, \
    CLIMB_CATEGORY, KM
from ...data.activity import active_stats, times_for_distance, hrz_stats, max_med_stats, max_mean_stats, \
    direction_stats
from ...data.climb import find_climbs, Climb, add_climb_stats
from ...data.frame import activity_statistics, present, statistics
from ...data.response import response_stats
from ...lib.log import log_current_exception
from ...sql import StatisticJournalFloat, Constant, StatisticJournalText
from ...stats.calculate.power import PowerCalculator

log = getLogger(__name__)


class ActivityCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, climb=None, **kargs):
        self.climb_ref = climb
        super().__init__(*args, **kargs)

    def _read_dataframe(self, s, ajournal):
        try:
            adf = activity_statistics(s, DISTANCE, ELEVATION, HEART_RATE, HR_ZONE, POWER_ESTIMATE,
                                      SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                      activity_journal=ajournal, with_timespan=True, check=False)
            start, finish = ajournal.start - dt.timedelta(hours=1), ajournal.finish + dt.timedelta(hours=1)
            sdf = statistics(s, FATIGUE_D_ANY, FITNESS_D_ANY, start=start, finish=finish,
                             owner=self.owner_in, check=False)
            return adf, sdf
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise

    def _calculate_stats(self, s, ajournal, df):
        adf, sdf = df
        stats, climbs = {}, None
        stats.update(active_stats(adf))
        stats.update(self.__average_power(s, ajournal, stats[ACTIVE_TIME]))
        stats.update(times_for_distance(adf))
        stats.update(hrz_stats(adf))
        stats.update(max_med_stats(adf))
        stats.update(max_mean_stats(adf))
        stats.update(direction_stats(adf))
        stats.update(response_stats(sdf))
        if present(adf, ELEVATION):
            params = Climb(**loads(Constant.get(s, self.climb_ref).at(s).value))
            climbs = list(find_climbs(adf, params=params))
            add_climb_stats(adf, climbs)
        return df, stats, climbs

    def __average_power(self, s, ajournal, active_time):
        # this doesn't fit nicely anywhere...
        energy = StatisticJournalFloat.at(s, ajournal.start, ENERGY_ESTIMATE, PowerCalculator, ajournal.activity_group)
        if energy and active_time:
            return {MEAN_POWER_ESTIMATE: 1000 * energy.value / active_time}
        else:
            return {MEAN_POWER_ESTIMATE: 0}

    def _copy_results(self, s, ajournal, loader, data):
        df, stats, climbs = data
        self.__copy(ajournal, loader, stats, ACTIVE_DISTANCE, KM, summaries(MAX, CNT, SUM, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, ACTIVE_TIME, S, summaries(MAX, SUM, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, ACTIVE_SPEED, KMH, summaries(MAX, AVG, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, MEAN_POWER_ESTIMATE, W, summaries(MAX, AVG, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, DIRECTION, DEG, None, ajournal.start)
        self.__copy(ajournal, loader, stats, ASPECT_RATIO, None, None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, MIN_KM_TIME_ANY, S, summaries(MIN, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, MED_KM_TIME_ANY, S, summaries(MIN, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, PERCENT_IN_Z_ANY, PC, None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, TIME_IN_Z_ANY, S, None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, MAX_MED_HR_M_ANY, BPM, summaries(MAX, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, MAX_MEAN_PE_M_ANY, W, summaries(MAX, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, _delta(FATIGUE_D_ANY), FF, summaries(MAX, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, _delta(FITNESS_D_ANY), FF, summaries(MAX, MSR), ajournal.start)
        if climbs:
            loader.add(TOTAL_CLIMB, M, summaries(MAX, MSR), ajournal.activity_group, ajournal,
                       sum(climb[CLIMB_ELEVATION] for climb in climbs), ajournal.start, StatisticJournalFloat)
            for climb in sorted(climbs, key=lambda climb: climb[TIME]):
                self.__copy(ajournal, loader, climb, CLIMB_ELEVATION, M, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_DISTANCE, KM, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_TIME, S, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_GRADIENT, PC, summaries(MAX, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_POWER, W, summaries(MAX, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_CATEGORY, None, None, climb[TIME],
                            type=StatisticJournalText)
        if stats:
            log.warning(f'Unsaved statistics: {list(stats.keys())}')

    def __copy_all(self, ajournal, loader, stats, pattern, units, summary, time, type=StatisticJournalFloat):
        for name in like(pattern, stats):
            self.__copy(ajournal, loader, stats, name, units, summary, time, type=type)

    def __copy(self, ajournal, loader, stats, name, units, summary, time, type=StatisticJournalFloat):
        try:
            loader.add(name, units, summary, ajournal.activity_group, ajournal, stats[name], time, type)
            del stats[name]
        except:
            log.warning(f'Failed to load {name}')
            log_current_exception(traceback=False)




from bokeh import tile_providers
from bokeh.io import output_file
from bokeh.layouts import row, column
from bokeh.models import PreText, Slider
from bokeh.plotting import show, figure

from ch2.data import *
from ch2.sql import Segment
from ch2.jupyter.decorator import template


@template
def define_segment(local_time):

    f'''
    # Define Segment: {local_time.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data for a ride containing the segment.
    '''
    s = session('-v2')
    df = activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, ELEVATION,
                             LATITUDE, LONGITUDE, local_time=local_time)
    df.dropna(inplace=True)
    df.describe()

    '''
    ### Select Segment

    Use the sliders to isolate the segment.
    
    You may need to play with both sliders before the map displays correctly 
    (todo - fix this, and also all the log messages).  
    '''
    TILE = tile_providers.get_provider(tile_providers.Vendors.STAMEN_TERRAIN)

    output_file(filename='/dev/null')
    width, height = 800, 800

    def modify_doc(doc):

        t1 = PreText(height=20, width=200)
        t2 = PreText(height=20, width=200)
        t3 = PreText(height=20, width=100)
        s1 = Slider(start=0, end=len(df), value=0, title='Start')
        s2 = Slider(start=1, end=len(df)-1, value=len(df), title='Length')
        map = figure(plot_width=width, plot_height=height, x_axis_type='mercator', y_axis_type='mercator',
                     match_aspect=True)
        map.add_tile(TILE)
        map.circle(x=SPHERICAL_MERCATOR_X, y=SPHERICAL_MERCATOR_Y, source=df, name='map')
        elevation = figure(plot_width=width, plot_height=height//10)
        elevation.line(x=DISTANCE, y=ELEVATION, source=df, name='elevation')
        c = column(row(s1, s2), row(t1, t2, t3), map, elevation)

        def mkplot(l1, l2):
            l2 = min(len(df)-1, l1+l2)
            t1.text = '%9.5f,%9.5f' % (df.iloc[l1]['Longitude'], df.iloc[l1]['Latitude'])
            t2.text = '%9.5f,%9.5f' % (df.iloc[l2]['Longitude'], df.iloc[l2]['Latitude'])
            t3.text = '%4.2fkm' % ((df.iloc[l2]['Distance'] - df.iloc[l1]['Distance']) / 1000)
            get_renderer(map, 'map').data_source.data = df[l1:l2]
            get_renderer(elevation, 'elevation').data_source.data = df[l1:l2]

        s1.on_change('value', lambda attr, old, new: mkplot(s1.value, s2.value))
        s2.on_change('value', lambda attr, old, new: mkplot(s1.value, s2.value))
        doc.add_root(c)

    show(modify_doc)

    '''
    ## Define Segment
    
    Replace the details below with your own from the plot above.
    The start and finish values are (lat, lon); the distance is in metres.
    
    Finally, uncomment the `s.add()` to add this to the database.
    '''

    segment = Segment(start=(-70.61813, -33.41536), finish=(-70.63340, -33.42655), distance=4400,
                      name='San Cristobal', description='Climb up San Cristobal in Parque Metropolitana')
    #s.add(segment)
    #s.commit()




from sqlalchemy import or_, desc, distinct
from sqlalchemy.orm import aliased
from sqlalchemy.sql.functions import min

from . import JournalDiary
from ...diary.model import text, link, optional_text
from ...lib.date import to_time, time_to_local_time
from ...sql import ActivityJournal, ActivitySimilarity, ActivityNearby

NEARBY_LINKS = 'nearby-links'


# todo - replace w to_local_date?
def _fmt_time(time):
    return to_time(time).strftime('%Y-%m-%d')


def fmt_nearby(aj, nb):
    return to_time(aj.start).strftime('%y-%m-%d') + ' %d%%' % (nb.similarity * 100)


class NearbyDiary(JournalDiary):

    def _read_schedule(self, s, date, schedule):
        yield from []

    @optional_text('Nearby', tag='nearbys')
    def _read_journal_date(self, s, ajournal, date):
        for constraint in constraints(s):
            results = list(self.__read_constraint(s, ajournal, constraint))
            if results: yield [text(constraint, tag='nearby')] + results

    def __read_constraint(self, s, ajournal, c):
        for title, callback, fmt in (('Any Time', nearby_any_time,
                                      lambda x: link(fmt_nearby(*x), db=(time_to_local_time(x[0].start),))),
                                     ('Earlier', nearby_earlier,
                                      lambda x: link(fmt_nearby(*x), db=(time_to_local_time(x[0].start),))),
                                     ('All', constraint,
                                      lambda x: link(_fmt_time(x.start), db=(time_to_local_time(x.start),)))):
            links = [fmt(result) for result in callback(s, ajournal, c)]
            if links:
                yield [text(title, tag=NEARBY_LINKS)] + links


def constraints(s):
    yield from (c[0] for c in
                s.query(distinct(ActivitySimilarity.constraint)).
                    order_by(ActivitySimilarity.constraint).all())


def constraint(s, ajournal, constraint):
    nb_us = aliased(ActivityNearby)
    aj_us = aliased(ActivityJournal)
    nb_them = aliased(ActivityNearby)
    aj_them = aliased(ActivityJournal)
    return [nb.activity_journal
            for nb in s.query(nb_them).
                join(nb_us, nb_us.constraint == nb_them.constraint).
                join(aj_them, aj_them.id == nb_them.activity_journal_id).
                join(aj_us, aj_us.id == nb_us.activity_journal_id).
                filter(nb_us.constraint == constraint,
                       nb_us.activity_journal == ajournal,
                       nb_them.activity_journal != ajournal,
                       aj_us.activity_group_id == aj_them.activity_group_id).
                order_by(aj_them.start)]


def nearby_earlier(s, ajournal, constraint=None, threshold=0.3):
    aj_lo = aliased(ActivityJournal)
    aj_hi = aliased(ActivityJournal)
    q = s.query(ActivitySimilarity). \
        join(aj_hi, ActivitySimilarity.activity_journal_hi_id == aj_hi.id). \
        join(aj_lo, ActivitySimilarity.activity_journal_lo_id == aj_lo.id). \
        filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                   ActivitySimilarity.activity_journal_lo_id == ajournal.id),
               aj_lo.activity_group_id == aj_hi.activity_group_id,
               or_(aj_hi.id == ajournal.id, aj_hi.start < ajournal.start),
               or_(aj_hi.id == ajournal.id, aj_lo.start < ajournal.start),
               ActivitySimilarity.similarity > threshold). \
        order_by(desc(min(aj_lo.start, aj_hi.start)))
    if constraint:
        q = q.filter(ActivitySimilarity.constraint == constraint)
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]


def nearby_any_time(s, ajournal, constraint=None, threshold=0.3):
    aj_lo = aliased(ActivityJournal)
    aj_hi = aliased(ActivityJournal)
    q = s.query(ActivitySimilarity). \
        join(aj_hi, ActivitySimilarity.activity_journal_hi_id == aj_hi.id). \
        join(aj_lo, ActivitySimilarity.activity_journal_lo_id == aj_lo.id). \
        filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                   ActivitySimilarity.activity_journal_lo_id == ajournal.id),
               aj_lo.activity_group_id == aj_hi.activity_group_id,
               ActivitySimilarity.similarity > threshold). \
        order_by(desc(ActivitySimilarity.similarity))
    if constraint:
        q = q.filter(ActivitySimilarity.constraint == constraint)
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]




from logging import getLogger

import numpy as np

from ch2.lib.image import xy_to_latlon, extract_rgb, create_image
from ..srtm.bilinear import bilinear_elevation_from_constant

log = getLogger(__name__)


def create_elevation(s, image):
    x, y = np.mgrid[0:image.height, 0:image.width]
    x, y = x.reshape(image.width * image.height), y.reshape(image.width * image.height)
    lat, lon = xy_to_latlon(image, x, y)
    oracle = bilinear_elevation_from_constant(s)
    elevation = np.array([oracle.elevation(lat, lon) for lat, lon in zip(lat, lon)])
    return elevation.reshape(image.height, image.width)


def add_elevation(image, elevation):
    r, g, b = extract_rgb(image)
    e = ((elevation / np.max(elevation)) * np.iinfo(r.dtype).max).astype(r.dtype)
    return create_image(image, np.stack([r, g, b, e]))




from abc import abstractmethod
from enum import IntEnum
from logging import getLogger

from sqlalchemy import ForeignKey, Column, Integer, func, UniqueConstraint, select, Boolean
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql.functions import count

from ..support import Base
from ..types import OpenSched, Date, ShortCls, short_cls
from ..utils import add
from ...lib.date import to_time, time_to_local_date, max_time, min_time, extend_range

log = getLogger(__name__)


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    DIARY_TOPIC = 3
    CONSTANT = 4
    MONITOR = 5
    SEGMENT = 6
    COMPOSITE = 7
    DUMMY = 8
    ITEM = 9
    MODEL = 10
    ACTIVITY_TOPIC = 11


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast deletes of subtypes

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @abstractmethod
    def time_range(self, s):
        raise NotImplementedError('time_range for %s' % self)

    @classmethod
    def before_flush(cls, s):
        cls.__clean_dirty_intervals(s)

    @classmethod
    def __clean_dirty_intervals(cls, s):
        from .statistic import StatisticJournal, StatisticJournalText
        # sessions are generally restricted to one time region, so we'll bracket that rather than list all times
        start, finish = None, None
        # all sources except intervals that are being deleted (need to catch on cascade to statistics)
        for instance in s.deleted:
            if isinstance(instance, Source) and not isinstance(instance, Interval):
                a, b = instance.time_range(s)
                start, finish = min_time(a, start), max_time(b, finish)
        # all modified statistics
        for instance in s.dirty:
            # ignore constants as time 0
            # ignore textual values (don't have useful stats and commonly modified fields)
            if isinstance(instance, StatisticJournal) and s.is_modified(instance) and instance.time \
                    and not isinstance(instance, StatisticJournalText):
                start, finish = extend_range(start, finish, instance.time)
        # all new statistics that aren't associated with intervals and have non-null data
        # (avoid triggering on empty diary entries)
        for instance in s.new:
            # ignore constants as time 0
            # ignore textual values
             if isinstance(instance, StatisticJournal) and not isinstance(instance.source, Interval) \
                     and not isinstance(instance.source, Dummy) \
                     and not isinstance(instance.source, StatisticJournalText) \
                     and instance.value is not None and instance.time:
                start, finish = extend_range(start, finish, instance.time)
        if start is not None:
            Interval.mark_dirty_times(s, start, finish)


@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    Source.before_flush(session)


class NoStatistics(Exception):
    pass


class Interval(Source):

    __tablename__ = 'interval'   # would be nice to rename this interval_source at some point

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False, index=True)
    # disambiguate creator so each can wipe only its own data on force
    owner = Column(ShortCls, nullable=False)
    # these are for the schedule - finish is redundant (start is not because of timezone issues)
    start = Column(Date, nullable=False, index=True)
    finish = Column(Date, nullable=False, index=True)
    dirty = Column(Boolean, nullable=False, default=False)
    UniqueConstraint(schedule, owner, start)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __str__(self):
        owner = self.owner if isinstance(self.owner, str) else short_cls(self.owner)
        return 'Interval "%s from %s" (owner %s)' % (self.schedule, self.start, owner)

    @classmethod
    def _raw_statistics_time_range(cls, s, statistics_owner=None):
        '''
        The time range over which statistics exist (optionally restricted by owner),
        ignoring constants at "time zero".  This is the first to the last time for
        any statistics - it pays no attention to gaps.
        '''
        from .statistic import StatisticJournal, StatisticName
        q = s.query(func.min(StatisticJournal.time), func.max(StatisticJournal.time)). \
            filter(StatisticJournal.time > to_time(2 * 24 * 60 * 60.0))
        if statistics_owner:
            q = q.join(StatisticName).filter(StatisticName.owner == statistics_owner)
        start, finish = q.one()   # skip entire first day because tz
        if start and finish:
            return start, finish
        else:
            raise NoStatistics('No statistics are currently defined')

    @classmethod
    def at(cls, s, schedule, interval_owner, start):
        '''
        The existing interval for a given start, owner, schedule.
        '''
        return s.query(Interval). \
            filter(Interval.start == start,
                   Interval.schedule == schedule,
                   Interval.owner == interval_owner).one_or_none()

    @classmethod
    def missing_dates(cls, s, schedule, interval_owner, statistic_owner=None, start=None, finish=None):
        '''
        Previous approach was way too complicated and not thread-safe.  Instead, just enumerate intervals and test.
        '''
        stats_start_time, stats_finish_time = cls._raw_statistics_time_range(s, statistic_owner)
        stats_start = time_to_local_date(stats_start_time)
        stats_finish = time_to_local_date(stats_finish_time)
        log.debug('Statistics (in general) exist %s - %s' % (stats_start, stats_finish))
        start = schedule.start_of_frame(start if start else stats_start)
        finish = finish if finish else schedule.next_frame(stats_finish)
        while start < finish:
            next = schedule.next_frame(start)
            if not cls.at(s, schedule, interval_owner, start):
                yield start, next
            start = next

    @classmethod
    def dirty_all(cls, s):
        log.warning('Dirtying all Intervals')
        s.query(Interval).update({Interval.dirty: True}, synchronize_session=False)

    @classmethod
    def mark_dirty_times(cls, s, start, finish, owner=None):
        '''
        Dirty all intervals that include data in the given TIME range,
        '''
        cls.mark_dirty_dates(s, time_to_local_date(start), time_to_local_date(finish), owner=owner)

    @classmethod
    def mark_dirty_dates(cls, s, start, finish, owner=None):
        '''
        Dirty all summary intervals (not monitor intervals) in the given DATE range.
        '''
        q = s.query(Interval).filter(Interval.start <= finish, Interval.finish > start)
        if owner:
            q = q.filter(Interval.owner == owner)
        q.update({Interval.dirty: True}, synchronize_session=False)

    @classmethod
    def clean(cls, s):
        q = s.query(Interval).filter(Interval.dirty == True)
        log.info(f'Deleting {q.count()} dirty Intervals')
        for interval in q.all():
            s.delete(interval)


class Dummy(Source):

    __tablename__ = 'dummy_source'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.DUMMY
    }

    def time_range(self, s):
        return None, None

    @classmethod
    def singletons(cls, s):
        from .. import StatisticName
        source = s.query(Dummy).one()
        name = s.query(StatisticName).filter(StatisticName.owner == source).one()
        return source, name


class CompositeComponent(Base):

    __tablename__ = 'composite_component'

    id = Column(Integer, primary_key=True)
    input_source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'))
    input_source = relationship('Source', foreign_keys=[input_source_id])
    output_source_id = Column(Integer, ForeignKey('composite_source.id', ondelete='cascade'))
    output_source = relationship('Composite', foreign_keys=[output_source_id])
    UniqueConstraint(output_source_id, input_source_id)  # ordered so output is a useful index


class Composite(Source):

    __tablename__ = 'composite_source'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    components = relationship('Source', secondary='composite_component')
    n_components = Column(Integer, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.COMPOSITE
    }

    def time_range(self, s):
        return None, None

    @classmethod
    def clean(cls, s):
        from ...data.frame import _tables
        t = _tables()
        id, target, actual = t.cmp.c.id.label("id"), t.cmp.c.n_components.label('target'), count(t.cc.c.id).label('actual')
        # had horrible bug here when join condition was filter (ie 'where') rather than a condition in the join
        # (ie 'on') - the where version doesn't expand to nulls and so the count is wrong when zero.
        q1 = select([id, target, actual]). \
             select_from(t.cmp.outerjoin(t.cc, t.cmp.c.id == t.cc.c.output_source_id)). \
             group_by(t.cmp.c.id)
        q2 = select([q1.c.id]).where(q1.c.target != q1.c.actual)
        q3 = select([count()]).select_from(q2)
        n = s.connection().execute(q3).scalar()
        if n:
            log.warning(f'Deleting Composite entries due to missing components')
            total = 0
            while n:
                q4 = t.src.delete().where(t.src.c.id.in_(q2.cte()))
                s.connection().execute(q4)
                total += n
                n = s.connection().execute(q3).scalar()
            log.warning(f'Deleted {total} Composite entries')
            s.commit()

    @classmethod
    def find(cls, s, *sources):
        from ...data.frame import _tables
        ids = [source.id if isinstance(source, Source) else source for source in sources]
        t = _tables()
        q = s.query(Composite)
        for id in ids:
            component = t.cc.alias()
            q = q.join(component, Composite.id == component.c.output_source_id).\
                filter(component.c.input_source_id == id)
        return q.one_or_none()

    @classmethod
    def create(cls, s, *sources):
        composite = add(s, Composite(n_components=0))
        for source in sources:
            add(s, CompositeComponent(input_source=source, output_source=composite))
            composite.n_components += 1
        return composite




import datetime as dt
from logging import getLogger

from urwid import MainLoop, connect_signal

from .args import DATE, SCHEDULE, FAST, mm, BASE
from ..diary.database import read_date, COMPARE_LINKS, read_schedule
from ..diary.views.urwid import build, layout_date, layout_schedule
from ..jupyter.server import set_controller, JupyterController
from ..jupyter.template.activity_details import activity_details
from ..jupyter.template.all_activities import all_activities
from ..jupyter.template.compare_activities import compare_activities
from ..jupyter.template.health import health
from ..jupyter.template.similar_activities import similar_activities
from ..lib.date import to_date, time_to_local_date, time_to_local_time, local_time_to_time
from ..lib.io import tui
from ..lib.schedule import Schedule
from ..lib.utils import PALETTE
from ..lib.widgets import DateSwitcher
from ..sql import PipelineType, DiaryTopicJournal, ActivityJournal, Interval
from ..sql.database import StatisticJournal
from ..stats.display.nearby import NEARBY_LINKS
from ..stats.pipeline import run_pipeline
from ..urwid.tui.factory import Factory
from ..urwid.tui.tabs import TabList

log = getLogger(__name__)


@tui
def diary(args, system, db):
    '''
## diary

    > ch2 diary [DATE]

The date can be an absolute day or the number of days previous.  So `ch2 diary 1` selects yesterday.

Display the daily diary.  Enter information here.

To exit, alt-q (or, without saving, alt-x).

    > ch2 diary (--month | --year | --schedule SCHEDULE) [DATE}

Display a summary for the month / year / schedule.
    '''
    set_controller(JupyterController(args, system))
    date, schedule, base = args[DATE], args[SCHEDULE], args[BASE]
    if not date:
        date = dt.date.today()
    else:
        days = None
        try:
            # try int first because we need to separate the case of days from years
            days = int(date)
            if days > 1000:
                days = None
        except ValueError:
            pass
        if days is None:
            date = to_date(date)
        else:
            date = dt.date.today() - dt.timedelta(days=days)
    with db.session_context() as s:
        DiaryTopicJournal.check_tz(system, s)
    if schedule:
        schedule = Schedule(schedule)
        if schedule.start or schedule.finish:
            raise Exception('Schedule must be open (no start or finish)')
        MainLoop(ScheduleDiary(db, date, schedule), palette=PALETTE).run()
    else:
        MainLoop(DailyDiary(db, date), palette=PALETTE).run()
        if not args[FAST]:
            print('\n  Please wait while statistics are updated...')
            run_pipeline(system, db, base, PipelineType.STATISTIC)
            print(f'  ...done (thanks! - use {mm(FAST)} to avoid this, if the risk is worth it)\n')


class Diary(DateSwitcher):

    def __init__(self, db, date):
        super().__init__(db, date)

    def save(self):
        s = self._session
        if s:
            self.__clean(s, s.dirty, delete=True)
            self.__clean(s, s.new, delete=False)
        super().save()

    @staticmethod
    def __clean(s, instances, delete=False):
        for instance in instances:
            if isinstance(instance, StatisticJournal):
                if instance.value == None:
                    log.debug(f'Discarding {instance}')
                    if delete:
                        s.delete(instance)
                    else:
                        s.expunge(instance)

    def _wire(self, active, name, callback):
        if name in active:
            for menu in active[name]:
                connect_signal(menu, 'click', callback)
            del active[name]


class DailyDiary(Diary):
    '''
    Render the diary at a given date.
    '''

    def _build(self, s):
        log.debug('Building diary at %s' % self._date)
        model = list(read_date(s, self._date))
        f = Factory(TabList())
        active, widget = build(model, f, layout_date)
        self._wire(active, NEARBY_LINKS, lambda m: self._change_date(to_date(m.state[0].split(' ')[0])))
        self._wire(active, COMPARE_LINKS, lambda m: self.__show_gui(*m.state))
        self._wire(active, 'health', lambda l: self.__show_health())
        self._wire(active, 'all-similar', lambda l: self.__show_similar(l.state[0]))
        if active:
            raise Exception(f'Unhandled links: {", ".join(active.keys())}')
        return widget, f.tabs

    def __show_gui(self, aj1, aj2, group):
        aj1t, aj2t = local_time_to_time(aj1), local_time_to_time(aj2) if aj2 else None
        aj1 = ActivityJournal.at_local_time(self._session, aj1)
        if aj2:
            compare_activities(aj1t, aj2t, aj1.activity_group.name)
        else:
            activity_details(aj1t, aj1.activity_group.name)

    def __show_similar(self, local_time):
        aj1 = ActivityJournal.at_local_time(self._session, local_time)
        similar_activities(aj1.start, aj1.activity_group.name)

    def __show_health(self):
        health()


class ScheduleDiary(Diary):
    '''
    Display summary data for the given schedule.
    '''

    def __init__(self, db, date, schedule):
        self._schedule = schedule
        super().__init__(db, schedule.start_of_frame(date))

    def _build(self, s):
        log.debug(f'Building diary at {self._date} for {self._schedule}')
        model = list(read_schedule(s, self._schedule, self._date))
        f = Factory(TabList())
        active, widget = build(model, f, layout_schedule)
        self._wire(active, 'all-activities', lambda l: self.__show_all())
        if active:
            raise Exception(f'Unhandled links: {", ".join(active.keys())}')
        return widget, f.tabs

    def __show_all(self):
        finish = self._schedule.next_frame(self._date)
        all_activities(self._date.strftime('%Y-%m-%d'), finish.strftime('%Y-%m-%d'))




from operator import add

from bokeh.models import GMapOptions
from bokeh.plotting import show, gmap

from ch2.data import *
from ch2.data.plot import DEFAULT_BACKEND
from ch2.jupyter.decorator import template
from ch2.lib import *
from ch2.sql import *


@template
def gmap_activities(start, finish, activity_group_name, google_key):

    f'''
    # Google Maps Activities: {start.split()[0]} - {finish.split()[0]} / {activity_group_name}
    '''

    '''
    $contents
    '''

    '''
    ## Read Data
    '''

    s = session('-v2')
    data_frames = [activity_statistics(s, LATITUDE, LONGITUDE, activity_journal=aj)
                   for aj in s.query(ActivityJournal).
                       filter(ActivityJournal.start >= local_date_to_time(start),
                              ActivityJournal.start < local_date_to_time(finish),
                              ActivityJournal.activity_group == ActivityGroup.from_name(s, activity_group_name)).
                       all()]
    data_frames = [data_frame.dropna() for data_frame in data_frames if not data_frame.dropna().empty]
    print(f'Found {len(data_frames)} activities')

    '''
    ## Calculate Centre
    '''

    ll = [(data_frame[LATITUDE].mean(), data_frame[LONGITUDE].mean()) for data_frame in data_frames]
    ll = list(zip(*ll))
    ll = (median(ll[0]), median(ll[1]))

    '''
    ## Display
    '''

    map_options = GMapOptions(lat=ll[0], lng=ll[1], map_type="roadmap", scale_control=True)
    f = gmap(google_key, map_options, title=f'{start.split()[0]} - {finish.split()[0]} / {activity_group_name}',
             tools='pan,zoom_in,zoom_out,reset,undo,redo,save', output_backend=DEFAULT_BACKEND)
    for data_frame in data_frames:
        f.line(x=LONGITUDE, y=LATITUDE, source=data_frame)

    show(f)



from json import dumps
from logging import getLogger

from . import name_constant, add_enum_constant, set_constant, add_statistics
from ..stats.calculate.power import Power, ExtendedPowerCalculator

log = getLogger(__name__)

POWER_ESTIMATE_CNAME = 'PowerEstimate'


def add_power_estimate(s, c, activity_group,
                       bike='${Constant:Power.${SegmentReader:kit}:None}',
                       rider_weight='${DiaryTopic:Weight:DiaryTopic \"Status\" (d)}',
                       vary='wind_speed, wind_heading, slope'):
    '''
    Add the constants necessary to estimate power output.
    '''
    log.debug(f'Adding power statistics for {activity_group.name}')
    activity_group_constraint = str(activity_group)
    power_name = name_constant(POWER_ESTIMATE_CNAME, activity_group)
    add_enum_constant(s, power_name, Power,
                      {'bike': bike, 'rider_weight': rider_weight, 'vary': vary},
                      single=False, constraint=activity_group_constraint, description='''
Parameters used in estimating power.

These are complex and reference other statistics.  
For example, ${SegmentReader:kit} is the kit specified when the activity is uploaded.

* Bike is an expression that expends to identify a constant named by the kit which itself contains CdA, Crr and weight data for the bike.
* Rider_weight is an expression that expends to read the weight from the diary.
* Vary is an experimental parameter to select what attributes of the ride are modelled (leave blank).
''')
    add_statistics(s, ExtendedPowerCalculator, c, owner_in='[unused - data via activity_statistics]',
                   power=name_constant(POWER_ESTIMATE_CNAME, activity_group),
                   activity_group_name=activity_group.name)




import datetime as dt
from abc import abstractmethod
from logging import getLogger

from sqlalchemy import desc
from urwid import ExitMainLoop

from ..sql.tables.activity import ActivityJournal
from .date import DAY, WEEK, MONTH, YEAR, to_time, to_date, add_date, local_date_to_time, time_to_local_date
from ..urwid.tui.tabs import TabNode

log = getLogger(__name__)


class App(TabNode):
    '''
    An urwid mainloop, database session and tabs.
    '''

    def __init__(self, db):
        self.__db = db
        self.__session = None
        super().__init__(*self._build(self.__new_session()))

    def __new_session(self):
        if self.__session:
            self.save()
            self.__session.close()
        self.__session = self.__db.session(autoflush=False)
        return self.__session

    @property
    def _session(self):
        return self.__session

    def keypress(self, size, key):
        if key == 'meta q':
            self.save()
            raise ExitMainLoop()
        elif key == 'meta x':
            self.abort()
            raise ExitMainLoop()
        elif key == 'meta s':
            self.save()
        else:
            return super().keypress(size, key)

    def abort(self):
        if self.__session:
            self.__session.rollback()

    def save(self):
        if self.__session:
            self.__session.commit()

    @abstractmethod
    def _build(self, session):
        pass

    def rebuild(self):
        widget, tabs = self._build(self.__new_session())
        self._w = widget
        try:
            self.replace(tabs)
        except Exception as e:
            log.warning('Could not replace tabs: %s' % e)


class DateSwitcher(App):
    '''
    Extend App with shortcuts for changing date and rebuilding.
    '''

    def __init__(self, db, date):
        self.__date = date
        super().__init__(db)

    def keypress(self, size, key):
        if key.startswith('meta'):
            c = key[-1]
            if c.lower() in (DAY, WEEK, MONTH, YEAR, 't'):
                self._next_prev_date(c)
                return
            if c.lower() == 'a':
                self._next_prev_activity(c)
                return
        return super().keypress(size, key)

    def _next_prev_activity(self, c):
        s = self._session
        if c == 'a':
            journal = ActivityJournal.before_local_time(s, self._date)
        else:
            journal = ActivityJournal.after_local_time(s, self._date)
        if journal:
            self._change_date(time_to_local_date(journal.start))

    def _next_prev_date(self, c):
        if c.lower() == 't':
            date = dt.date.today()
        else:
            delta = (-1 if c == c.lower() else 1, c.lower())
            date = add_date(self.__date, delta)
        self._change_date(date)

    def _change_date(self, date):
        self.__date = self._refine_new_date(date)
        self.rebuild()

    def _refine_new_date(self, date):
        '''
        Hook for subclasses
        '''
        return date

    @property
    def _date(self):
        return self.__date




from collections import defaultdict
from logging import getLogger
from time import sleep

from sqlalchemy.exc import IntegrityError

from .waypoint import make_waypoint
from ..commands.args import UNLOCK
from ..sql import StatisticJournal, StatisticName, Dummy, Interval
from ..sql.tables.statistic import STATISTIC_JOURNAL_CLASSES, STATISTIC_JOURNAL_TYPES
from ..sql.types import short_cls

log = getLogger(__name__)


class StatisticJournalLoader:

    # we want to load multiple tables (because we're using inheritance) quickly and safely.
    # to do this, we need to generate IDs ourselves (to avoid reading the ID for the base class)
    # on sqlite we can do this by:
    # - using a single transaction
    # - starting to write (a dummy StatisticJournal, which also gets us the last ID)
    # - calculating IDs from the dummy ID
    # - writing the data (still in the same transaction)
    # - removing the dummy
    # - committing
    # afaict that would work on any SQL database with a reasonable level of isolation.
    # what is special to sqlite is that the final commit will not fail, because there is only ever one
    # process writing (this is true even when using multiple processes)

    # well the above worked for a while. then started throwing exceptions, so i needed to add
    # the while loop below.

    def __init__(self, s, owner, add_serial=True, clear_timestamp=True, abort_after=100):
        self._s = s
        self._owner = owner
        self.__statistic_name_cache = dict()
        self.__staging = defaultdict(list)
        self.__latest = dict()
        self.__add_serial = add_serial
        self.__start = None
        self.__finish = None
        self.__last_time = None
        self.__serial = 0 if add_serial else None
        self.__clear_timestamp = clear_timestamp
        self.__abort_after = abort_after
        self.__counts = defaultdict(lambda: 0)

    @property
    def start(self):
        return self.__start

    @property
    def finish(self):
        return self.__finish

    def __bool__(self):
        return bool(self.__staging)

    def load(self):
        self._s.commit()
        dummy = self._preload()
        try:
            self._load_ids(dummy)
        except Exception:
            self._s.rollback()
            # dummy may have been deleted in the rollback - it depends if there were any intermediate commits
            # (which is a whole other problem), so to be sure...
            self.unlock(self._s)
            raise
        self._postload()

    def _preload(self):
        dummy_source, dummy_name = Dummy.singletons(self._s)
        dummy, count = None, 0
        while not dummy:
            try:
                log.debug(f'Trying to acquire database ({count})')
                dummy = StatisticJournal(source=dummy_source, statistic_name=dummy_name, time=0.0)
                self._s.add(dummy)
                self._s.flush()
                log.debug('Acquired database')
            except IntegrityError:
                log.debug('Failed to acquire database')
                self._s.rollback()
                dummy, count = None, count+1
                if count > self.__abort_after:
                    raise Exception(f'Could not acquire database after {count} attempts '
                                    f'(you may need to use `ch2 {UNLOCK}` once all workers have stopped)')
                sleep(0.1)
        log.debug(f'Dummy ID {dummy.id}')
        return dummy

    def _load_ids(self, dummy):
        rowid, count = dummy.id + 1, 0
        for type in self.__staging:
            log.debug('Loading %d values for type %s' % (len(self.__staging[type]), short_cls(type)))
            for i in range(min(5, len(self.__staging[type]))):
                sjournal = self.__staging[type][i]
                log.debug(f'Example: {sjournal.value} at {sjournal.time}')
            for sjournal in self.__staging[type]:
                sjournal.id = rowid
                rowid += 1
            self._s.bulk_save_objects(self.__staging[type])
            count += len(self.__staging[type])
        self._s.commit()
        log.info(f'Loaded {count} statistics')
        log.debug('Removing Dummy')
        self._s.delete(dummy)
        self._s.commit()
        log.debug('Dummy removed')

    def _postload(self):
        # manually clean out intervals because we're doing a fast load
        if self.__clear_timestamp and self.start and self.finish:
            Interval.mark_dirty_times(self._s, self.start, self.finish)
            self._s.commit()

    @classmethod
    def unlock(cls, s):
        dummy_source, dummy_name = Dummy.singletons(s)
        s.query(StatisticJournal). \
            filter(StatisticJournal.source == dummy_source,
                   StatisticJournal.statistic_name == dummy_name).delete()
        s.commit()

    def add(self, name, units, summary, constraint, source, value, time, cls):

        if self.__add_serial:
            if self.__last_time is None:
                self.__last_time = time
            elif time > self.__last_time:
                self.__last_time = time
                self.__serial += 1
            elif time < self.__last_time:
                raise Exception('Time travel - timestamp for statistic decreased')

        self.__start = min(self.__start, time) if self.__start else time
        self.__finish = max(self.__finish, time) if self.__finish else time

        key = (name, constraint)
        if key not in self.__statistic_name_cache:
            self.__statistic_name_cache[key] = \
                StatisticName.add_if_missing(self._s, name, STATISTIC_JOURNAL_TYPES[cls],
                                             units, summary, self._owner, constraint)

        try:
            source = source.id
        except AttributeError:
            pass  # literal id
        statistic_name = self.__statistic_name_cache[key]
        journal_class = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]
        if cls != journal_class:
            raise Exception(f'Inconsistent class for {name}: {cls}/{journal_class}')
        instance = journal_class(statistic_name_id=statistic_name.id, source_id=source, value=value,
                                 time=time, serial=self.__serial)
        if key in self.__latest:
            prev = self.__latest[key]
            if instance.time > prev.time:
                self.__latest[key] = instance
                self.__staging[journal_class].append(instance)
                self.__counts[key] += 1
            elif instance.time == prev.time:
                if instance.value == prev.value:
                    log.warning(f'Skipping duplicate for {name}')
                else:
                    self._resolve_duplicate(name, instance, prev)
            else:
                self.__staging[journal_class].append(instance)
                self.__counts[key] += 1
        else:
            self.__latest[key] = instance
            self.__staging[journal_class].append(instance)
            self.__counts[key] += 1

    def _resolve_duplicate(self, name, instance, prev):
        raise Exception(f'Duplicate time ({prev.time}) for {name} ({instance.value}/{prev.value})')

    def latest(self, name, constraint):
        return self.__latest.get((name, constraint))

    def as_waypoints(self, names):
        Waypoint = make_waypoint(names.values())
        time_to_waypoint = defaultdict(lambda: Waypoint())
        statistic_ids = dict((name.id, name) for name in self.__statistic_name_cache.values())
        for type in self.__staging:
            for sjournal in self.__staging[type]:
                name = statistic_ids[sjournal.statistic_name_id].name
                if name in names:
                    time_to_waypoint[sjournal.time] = \
                        time_to_waypoint[sjournal.time]._replace(**{'time': sjournal.time,
                                                                    names[name]: sjournal.value})
        return [time_to_waypoint[time] for time in sorted(time_to_waypoint.keys())]

    def coverage_percentages(self):
        total = max(self.__counts.values())
        for key, count in self.__counts.items():
            yield key, 100 * count / total



from tempfile import TemporaryDirectory
from tests import LogTestCase

from ch2 import activities
from ch2.commands.args import bootstrap_dir, mm, DEV, FAST, V, m
from ch2.config import default, getLogger
from ch2.diary.database import read_date
from ch2.diary.model import LABEL, VALUE
from ch2.lib import to_date

log = getLogger(__name__)


class TestPower(LogTestCase):

    def test_constant(self):
        with TemporaryDirectory() as f:
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV),
                                           'activities', 'data/test/source/personal/2018-03-04-qdp.fit',
                                           '-Kn_cpu=1')
            activities(args, sys, db)
            with db.session_context() as s:
                model = list(read_date(s, to_date('2018-03-04')))
                print(model, len(model))
                [title, diary, achievements, activity, jupyter, database] = model
                print(activity)
                name = activity[1]
                print(name)
                self.assertEqual(name[LABEL], 'Name')
                self.assertEqual(name[VALUE], '2018-03-04-qdp')
                route = activity[2]
                self.assertEqual(route[LABEL], 'Route')




from itertools import groupby
from logging import getLogger
from math import sqrt, ceil

from pygeotile.point import Point

from .activity import ActivityReader
from ..waypoint import filter_none
from ...rtree import MatchType
from ...rtree.spherical import LocalTangent, SQRTree, Global
from ...lib.date import to_time, format_time
from ...lib.utils import sign
from ...sql.database import Timestamp
from ...sql.tables.activity import ActivityGroup
from ...sql.tables.segment import Segment, SegmentJournal
from ...sql.utils import add

log = getLogger(__name__)

NAMES = {'Latitude': 'lat',
         'Longitude': 'lon',
         'Distance': 'distance'}


class CalcFailed(Exception): pass


class SegmentReader(ActivityReader):

    def __init__(self, *args, inner_bound=10, match_bound=25, **kargs):
        self.inner_bound = inner_bound
        self.match_bound = match_bound
        super().__init__(*args, **kargs)

    def _startup(self, s):
        SegmentJournal.clean(s)
        super()._startup(s)
        self.__segments = {}
        for agroup in s.query(ActivityGroup).all():
            self.__segments[agroup.id] = self._read_segments(s, agroup)

    def _shutdown(self, s):
        SegmentJournal.clean(s)
        super()._shutdown(s)

    def _load_data(self, s, loader, data):
        super()._load_data(s, loader, data)
        ajournal, activity_group, first_timestamp, file_scan, define, records = data
        self._find_segments(s, ajournal, filter_none(NAMES.values(), loader.as_waypoints(NAMES)))

    def _find_segments(self, s, ajournal, waypoints):
        matches = sorted(self._initial_matches(s, ajournal.activity_group_id, waypoints), key=lambda m: m[2].name)
        for segment, segment_matches in groupby(matches, key=lambda m: m[2]):
            log.debug(f'Considering {segment.name}')
            ordered = sorted(segment_matches, key=lambda m: m[0])
            coallesced = list(self._coallesce(ordered))
            starts, finishes, segment = self._split(coallesced)
            while starts:
                start = starts.pop(-1)  # work backwards through starts
                copy = list(finishes)
                while copy:
                    finish = copy.pop(0)  # work forwards through finishes
                    if finish[0] > start[0]:
                        if self._try_segment(s, start, finish, waypoints, segment, ajournal):
                            copy = None  # exit search for this start
                            # move to a start that won't overlap (working backwards)
                            while starts and starts[-1][0] > start[0] - 0.5 * (finish[0] - start[0]):
                                starts.pop(-1)
                            if starts:
                                log.debug('Possible second segment')

    def _try_segment(self, s, starts, finishes, waypoints, segment, ajournal):
        try:
            log.info('Trying segment %s-%s for %s' % (starts, finishes, segment.name))
            d = waypoints[self._mid(finishes)].distance - waypoints[self._mid(starts)].distance
            if abs(d - segment.distance) / segment.distance > 0.2:
                raise CalcFailed('Distance between start and finish (%.1fkm) doesn\'t match segment (%.1fkm)' %
                                 (d / 1000, segment.distance / 1000))
            start_time = self._end_point(starts, waypoints, segment.start, self.inner_bound, True)
            finish_time = self._end_point(finishes, waypoints, segment.finish, self.inner_bound, False)
            sjournal = add(s, SegmentJournal(segment_id=segment.id, activity_journal=ajournal,
                                             start=start_time, finish=finish_time))
            log.info('Added %s for %s - %s' %
                           (segment.name, format_time(start_time), format_time(finish_time)))
            s.commit()  # needed to get id on sjournal (and let other workers in)
            # since source is sjournal we cannot use on_success
            Timestamp.set(s, self, constraint=segment, source=sjournal)
            return True
        except CalcFailed as e:
            log.warning(str(e))
            return False

    def _mid(self, indices):
        n = len(indices)
        return indices[n // 2]

    def _end_point(self, indices, waypoints, p, inner, hi_to_lo):
        '''
        Find the time of the point that makes the segment shortest while also being at a local
        minimum in distance from the start/finish point that is within inner m.
        '''
        metric = LocalTangent(p)
        lo, hi = self._limits(metric, waypoints, indices, p)
        log.info('Finding closest point to %s within %d: %f, %d: %f' %
                       (p, lo, self._dw(metric, p, waypoints, lo),
                        hi, self._dw(metric, p, waypoints, hi)))
        if hi_to_lo:
            start, finish = hi, lo
        else:
            start, finish = lo, hi
        minimum = self._minimum(metric, waypoints, start, finish, p, inner)
        # interpolate to fractional index
        i0, i1 = int(minimum), int(minimum) + 1
        k = (minimum - i0) / (i1 - i0)
        t0, t1 = waypoints[i0].time.timestamp(), waypoints[i1].time.timestamp()
        return to_time(t0 * (1 - k) + t1 * k)

    def _plot(self, p):
        '''
        Plot points for debugging
        '''
        lon, lat = p
        print('PLOT: %s' % (Point.from_latitude_longitude(lat, lon).meters, ))

    def _plotw(self, metric, waypoints, i, d):
        '''
        Plot points for debugging
        '''
        i1, i2 = int(i), int(i) + 1
        x1, y1 = metric.normalize((waypoints[i1].lon, waypoints[i1].lat))
        x2, y2 = metric.normalize((waypoints[i2].lon, waypoints[i2].lat))
        k = i - i1
        xk = x1 * (1 - k) + x2 * k
        yk = y1 * (1 - k) + y2 * k
        lon, lat = metric.denormalize((xk, yk))
        print('PLOT: %s %.1f' % (Point.from_latitude_longitude(lat, lon).meters, d))

    def _minimum(self, metric, waypoints, start, finish, p, inner):
        '''
        Find the (fractional) index of the point with minimum distance to the endpoint within the two limits.
        '''
        # self._plot(p)
        i = start
        while True:
            i, min_d = self._next_local_minimum(metric, p, waypoints, i, start, finish)
            log.info('Local minimum %.1f: %f (< %.1f?)' % (i, min_d, inner))
            # self._plotw(metric, waypoints, i, min_d)
            if min_d < inner:
                return i
            # move to next segment
            if start < finish:
                i = int(i + 1)
            else:
                i = ceil(i - 1)

    def _next_local_minimum(self, metric, p, waypoints, i, start, finish):
        '''
        Find the next point with a minimum distance to p, along the waypoints (interpolated),
        starting from i and between start and finish (moving towards finish).
        '''
        # this from basic algebra (derivation too long for a comment but pretty simple)
        # p0 is the point we want to minimize distance to
        # pi and pj are the end points of a nearby line segment
        # k is the fractional distance between pi and pj for the point nearest to p0
        # p1 is the point at k if it is between pi and pj
        # things are complicated by the need to return pj if it's the nearest point (discontinuity)
        p0 = metric.normalize(p)
        pi = metric.normalize((waypoints[i].lon, waypoints[i].lat))
        prev_discontinuity = None
        while start <= i < finish or start >= i > finish:
            j = i + sign(finish - start)
            pj = metric.normalize((waypoints[j].lon, waypoints[j].lat))
            dxji, dyji = pj[0] - pi[0], pj[1] - pi[1]
            dx0i, dy0i = p0[0] - pi[0], p0[1] - pi[1]
            k = (dxji * dx0i + dyji * dy0i) / (dxji**2 + dyji**2)
            if 0 < k < 1:  # we have a solution within the segment
                p1 = (pi[0] + k * (pj[0] - pi[0]), pi[1] + k * (pj[1] - pi[1]))
                return i + k * sign(finish - start), sqrt((p0[0] - p1[0])**2 + (p0[1] - p1[1])**2)
            # are we moving towards the target?
            if k > 1:
                # if so, save this endpoint as a possible minimum
                prev_discontinuity = (j, self._dw(metric, p, waypoints, j))
            else:
                # if not, then if we were before it was a minimum
                if prev_discontinuity:
                    return prev_discontinuity
            i, pi = j, pj
        raise CalcFailed('No minimum found')

    def _limits(self, metric, waypoints, indices, p):
        '''
        Given a pair of indices, move them apart so that the distance from the endpoint has increased
        in both directions (ie they bracket the endpoint in some sense).
        '''
        lo, hi = indices
        if lo == hi:
            lo, hi = lo-1, hi+1
        dl, dh = self._dw(metric, p, waypoints, lo), self._dw(metric, p, waypoints, hi)
        if dl < dh:
            lo, dl = self._inc_limit(metric, waypoints, p, dl, dh, lo, -1)
            hi, dh = self._inc_limit(metric, waypoints, p, dh, dl, hi, 1)
        else:
            hi, dh = self._inc_limit(metric, waypoints, p, dh, dl, hi, 1)
            lo, dl = self._inc_limit(metric, waypoints, p, dl, dh, lo, -1)
        log.info('Expanded %s to %s' % (indices, (lo, hi)))
        return lo, hi

    def _inc_limit(self, metric, waypoints, p, da, db, a, inc):
        '''
        Move one index.
        '''
        while da < db and ((inc > 0 and a < len(waypoints) - 1) or (inc < 0 and a > 0)):
            a += inc
            da = self._dw(metric, p, waypoints, a)
        return a, da

    def _dw(self, metric, p, waypoints, i):
        '''
        Distance between an waypoint and the endpoint.
        '''
        return self._d(metric, p, (waypoints[i].lon, waypoints[i].lat))

    def _d(self, metric, p1, p2):
        '''
        The distance between two points using the given metric.
        '''
        x1, y1 = metric.normalize(p1)
        x2, y2 = metric.normalize(p2)
        return sqrt((x1 - x2)**2 + (y1 - y2)**2)

    def _split(self, coallesced):
        '''
        Split waypoints for a single segment into separate start and finish points.
        '''
        starts, finishes, segment = [], [], None
        for i, start, segment in coallesced:
            if start:
                starts.append(i)
            else:
                finishes.append(i)
        return starts, finishes, segment

    def _coallesce(self, ordered_sm):
        '''
        Replace contiguous ranges of waypoints with a pair (min, max)
        '''
        prev, first = None, None
        for match in ordered_sm:
            if match != prev:  # skip duplicates
                if prev and (prev[1:2] != match[1:2] or prev[0]+1 != match[0]):
                    yield (first, prev[0]), prev[1], prev[2]
                    prev, first = None, None
                if first is None:
                    first = match[0]
            prev = match
        yield (first, prev[0]), prev[1], prev[2]

    def _initial_matches(self, s, agroup_id, waypoints):
        '''
        Check each waypoint against the r-tree and return all matches.
        '''
        found = set()
        for i, waypoint in enumerate(waypoints):
            for start, segment in self.__segments[agroup_id][[(waypoint.lon, waypoint.lat)]]:
                if segment not in found:
                    log.info(f'Candidate segment "{segment.name}" at ({segment.start_lon}, {segment.start_lat}) - '
                             f'({segment.finish_lon}, {segment.finish_lat})')
                    found.add(segment)
                log.debug(f'Match at {(waypoint.lon, waypoint.lat)} '
                          f'for {segment.name} {"start" if start else "finish"}')
                yield i, start, segment

    def _read_segments(self, s, agroup):
        '''
        Read segment endpoints into a global R-tree so we can detect when waypoints pass nearby.
        '''
        segments = Global(tree=lambda: SQRTree(default_border=self.match_bound, default_match=MatchType.OVERLAP))
        for segment in s.query(Segment).all():
            segments[[segment.start]] = (True, segment)
            segments[[segment.finish]] = (False, segment)
        if not segments:
            log.warning('No segments defined in database for %s' % agroup)
        return segments
